import logging
import re

from prestans import __version__
from prestans import deserializer
from prestans import exception
from prestans.rest import BlueprintHandler
from prestans.rest import ErrorResponse
from prestans.rest import DictionaryResponse
from prestans.rest import Request
from prestans.rest import Response
from prestans import serializer


class RequestRouter(object):
    """
    RequestRouter is a specialised WSGI router that primarily maps URLs to Handlers.
    All registered end-points must inherit from RequestHandler.

    RequestRouter sets the most likely response format based on Accept Headers. If
    no supported response format is found; RequestRouter sends back an HTML error.

    If the requested URL is not handled with the API; RequestRouter presents the
    client with a standardised error message.
    """

    def __init__(self, routes, serializers=None, default_serializer=None, deserializers=None,
                 default_deserializer=None, charset="utf-8", application_name="prestans",
                 logger=None, debug=False, description=None):

        self._application_name = application_name
        self._debug = debug
        self._routes = routes
        self._charset = charset
        self._description = description

        # are formats prestans handlers can send data back as
        self._serializers = serializers
        self._default_serializer = default_serializer
        # are formats prestans handlers can accept data as
        self._deserializers = deserializers
        self._default_deserializer = default_deserializer

        # init the default logger if one's not provided, this allows users to configure their own
        # http://www.blog.pythonlibrary.org/2012/08/02/python-101-an-intro-to-logging/
        if logger is None:
            logging.basicConfig()
            self._logger = logging.getLogger("prestans.%s" % application_name)
        else:
            self._logger = logger

        # set logger level, API can override this
        if self._debug:
            self._logger.setLevel(logging.DEBUG)
        else:
            self._logger.setLevel(logging.ERROR)

        # if serializers and deserializers aren't provided, prestans runs as a JSON app
        if serializers is None:
            self._serializers = [serializer.JSON()]

        if default_serializer is None:
            self._default_serializer = serializer.JSON()

        # deserializers
        if deserializers is None:
            self._deserializers = [deserializer.JSON()]

        if default_deserializer is None:
            self._default_deserializer = deserializer.JSON()

    @property
    def logger(self):
        return self._logger

    @property
    def description(self):
        return self._description

    @property
    def charset(self):
        return self._charset

    @property
    def debug(self):
        return self._debug

    @property
    def application_name(self):
        return self._application_name

    def __call__(self, environ, start_response):

        # say hello
        self.logger.info("%s exposes %i end-points; prestans %s; charset %s; debug %s" % (
            self._application_name, len(self._routes), __version__,
            self._charset, self._debug
        ))

        # validate serializers and deserializers; are subclasses of prestans.serializer.Base
        _default_outgoing_mime_types = list()
        for available_serializer in self._serializers:

            if not isinstance(available_serializer, serializer.Base):
                msg = "registered serializer %s.%s does not inherit from prestans.serializer.Serializer" % (
                    available_serializer.__module__,
                    available_serializer.__class__.__name__
                )
                raise TypeError(msg)

            _default_outgoing_mime_types.append(available_serializer.content_type())

        _default_incoming_mime_types = list()
        for available_deserializer in self._deserializers:

            if not isinstance(available_deserializer, deserializer.Base):
                msg = "registered deserializer %s.%s does not inherit from prestans.serializer.DeSerializer" % (
                    available_deserializer.__module__,
                    available_deserializer.__class__.__name__
                )
                raise TypeError(msg)

            _default_incoming_mime_types.append(available_deserializer.content_type())

        # report on the acceptable mime types
        self._logger.info("generally accepts %s; speaks %s" % (
            str(_default_outgoing_mime_types).strip("[]'"),
            str(_default_incoming_mime_types).strip("[]'")
        ))

        # attempt to parse the HTTP request
        request = Request(
            environ=environ,
            charset=self._charset,
            logger=self._logger,
            deserializers=self._deserializers,
            default_deserializer=self._default_deserializer
        )

        # initialise the route map
        route_map = self.generate_route_map(self._routes)

        try:

            # check if the requested URL has a valid registered handler
            for regexp, handler_class in route_map:

                # if absent, can assume to be empty string
                # https://www.python.org/dev/peps/pep-3333/#environ-variables
                match = regexp.match(environ.get("PATH_INFO", ""))

                # if we've found a match; ensure its a handler subclass and return it's callable
                if match:

                    # assemble the args and kwargs
                    args = match.groups()
                    kwargs = {}
                    for key, value in iter(regexp.groupindex.items()):
                        kwargs[key] = args[value - 1]

                    if len(kwargs) > 0:
                        args = ()

                    if issubclass(handler_class, BlueprintHandler):

                        response = DictionaryResponse(
                            charset=self._charset, logger=self._logger,
                            serializers=self._serializers,
                            default_serializer=self._default_deserializer
                        )

                        request_handler = handler_class(
                            args=args,
                            kwargs=kwargs,
                            request=request,
                            response=response,
                            logger=self._logger,
                            debug=self._debug,
                            route_map=self._routes
                        )
                    else:
                        response = Response(
                            charset=self._charset,
                            logger=self._logger,
                            serializers=self._serializers,
                            default_serializer=self._default_deserializer
                        )
                        response.minify = request.is_minified

                        request_handler = handler_class(
                            args=args,
                            kwargs=kwargs,
                            request=request,
                            response=response,
                            logger=self._logger,
                            debug=self._debug
                        )

                    return request_handler(environ, start_response)

            # request does not have a matched handler
            no_endpoint = exception.NoEndpointError()
            no_endpoint.request = request
            raise no_endpoint

        except exception.Base as exp:
            self.logger.error(exp)
            error_response = ErrorResponse(exp, self._default_serializer)
            return error_response(environ, start_response)

    @classmethod
    def generate_route_map(cls, routes):

        parsed_handler_map = []

        for url, handler in routes:

            regexp = url

            # patch regular expression if it is incomplete
            if not regexp.startswith('^'):
                regexp = '^' + regexp
            if not regexp.endswith('$'):
                regexp += '$'

            compiled_regex = re.compile(regexp)

            arg_count = compiled_regex.groups
            kwarg_count = len(compiled_regex.groupindex.items())

            if arg_count != kwarg_count and kwarg_count > 0:
                raise ValueError("%s URL is invalid, cannot mix named and un-named groups" % url)
            else:
                parsed_handler_map.append((compiled_regex, handler))

        return parsed_handler_map



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from datetime import date
from datetime import date as date_type
from datetime import datetime

from prestans import exception
from prestans.types import DataStructure


class Date(DataStructure):

    DEFAULT_FORMAT = "%Y-%m-%d"
    TODAY = '_PRESTANS_CONSTANT_MODEL_DATE_TODAY'

    def __init__(self, default=None, required=True, format=DEFAULT_FORMAT, description=None):

        if isinstance(default, date_type) or \
                default == self.TODAY or \
                default is None:
            self._default = default
        else:
            raise TypeError("default must be one of date, Date.TODAY")

        self._required = required
        self._format = format
        self._description = description

    @property
    def default(self):
        return self._default

    @property
    def required(self):
        return self._required

    @property
    def format(self):
        return self._format

    @property
    def description(self):
        return self._description

    def blueprint(self):

        blueprint = dict()
        blueprint['type'] = 'date'

        constraints = dict()
        constraints['default'] = self._default
        constraints['required'] = self._required
        constraints['format'] = self._format
        constraints['description'] = self._description

        blueprint['constraints'] = constraints
        return blueprint

    def validate(self, value):

        _validated_value = None

        if self._required and self._default is None and value is None:
            raise exception.RequiredAttributeError()
        elif self._required and value is None:
            if self._default == Date.TODAY:
                value = date.today()
            else:
                value = self._default
        elif not self._required and self._default is None and value is None:
            return _validated_value
        elif not self._required and value is None:
            if self._default == Date.TODAY:
                value = date.today()
            else:
                value = self._default

        from prestans.util import string_types

        if isinstance(value, date_type):
            _validated_value = value
        elif isinstance(value, string_types):
            try:
                _validated_value = datetime.strptime(value, self._format).date()
            except ValueError as exp:
                raise exception.ParseFailedError("date parsing failed %s" % exp)
        else:
            raise exception.ParseFailedError("cannot parse value of type %s" % value.__class__.__name__)

        return _validated_value

    def as_serializable(self, value):

        if not isinstance(value, date_type):
            raise exception.InvalidTypeError(value, 'datetime.date')

        return value.strftime(self._format)





class AttributeFilterImmutable(object):

    def __init__(self, attribute_filter):
        """
        Creates an immutable attribute filter based on a given mutable one.
        """
        self._key_map = dict()
        self._visible_keys = set()

        self._populate_from_filter(attribute_filter)

    def _populate_from_filter(self, attribute_filter):
        for key in attribute_filter.keys():
            if attribute_filter.is_filter_at_key(key):
                self._key_map[key] = AttributeFilterImmutable(getattr(attribute_filter, key))
                if self._key_map[key].are_any_attributes_visible():
                    self._visible_keys.add(key)
            else:
                self._key_map[key] = attribute_filter.is_attribute_visible(key)

                if self._key_map[key] is True:
                    self._visible_keys.add(key)

    def keys(self):
        return sorted(self._key_map.keys())

    def __contains__(self, key):
        return key in self._key_map

    def is_filter_at_key(self, key):
        """
        return True if attribute is a sub filter
        """
        return isinstance(self._key_map[key], self.__class__)

    def is_attribute_visible(self, key):
        """
        Returns True if an attribute is visible
        If attribute is an instance of AttributeFilter, it returns True if all attributes
        of the sub filter are visible.

        :param key: name of attribute to check
        :type key: str
        :return: whether attribute is visible
        :rtype: bool
        """
        return key in self._visible_keys

    def are_any_attributes_visible(self):
        """
        checks to see if any attributes are set to true
        """
        return len(self._visible_keys) > 0

    def are_all_attributes_visible(self):
        """
        checks to see if all attributes are set to true
        """
        return len(self._key_map) == len(self._visible_keys)

    def as_dict(self):
        """
        turns attribute filter object into python dictionary
        """
        output_dictionary = dict()

        for key, value in iter(self._key_map.items()):
            if isinstance(value, bool):
                output_dictionary[key] = value
            elif isinstance(value, self.__class__):
                output_dictionary[key] = value.as_dict()

        return output_dictionary

    def __getattr__(self, key):

        value = self._key_map.get(key)

        if value is not None:
            return value
        else:
            return super(AttributeFilterImmutable, self).__getattribute__(key)



import sys
import webob

from prestans import exception


class ErrorResponse(webob.Response):
    """
    ErrorResponse is a specialised webob.Response, it is responsible for writing
    out a message in the following format; using the currently selected serializer

      {
          "code": 404,
          "message": "This is an error message",
          "trace": [
            {
                "key": "value"
            }
          ]
      }
    """

    def __init__(self, raised_exception, serializer):

        super(ErrorResponse, self).__init__()

        self._exception = raised_exception
        self._serializer = serializer
        self._message = raised_exception.message
        self._stack_trace = raised_exception.stack_trace
        # self._trace = None

        # IETF hash dropped the X- prefix for custom headers
        # http://stackoverflow.com/q/3561381
        # http://tools.ietf.org/html/draft-saintandre-xdash-00

        from prestans import __version__ as version
        if not isinstance(version, str):
            version = version.encode("latin1")
        self.headers.add('Prestans-Version', version)

        self.content_type = self._serializer.content_type()
        self.status = raised_exception.http_status

    # @property
    # def trace(self):
    #     return self._trace
    #
    # def append_to_trace(self, trace_entry):
    #     """
    #     Use this to append to the stack trace
    #     """
    #     self._trace.append(trace_entry)

    def __call__(self, environ, start_response):

        # we have received a custom error response model, use it instead
        if isinstance(self._exception, exception.ResponseException) and self._exception.response_model:
            body_as_string = self._serializer.dumps(self._exception.response_model.as_serializable())
        # pack into default format for error response
        else:
            error_dict = {
                "code": self.status_int,
                "message": self._message,
                "trace": self._stack_trace
            }

            body_as_string = self._serializer.dumps(error_dict)

        self.content_length = len(body_as_string)

        start_response(self.status, self.headerlist)

        return [body_as_string.encode("utf-8")]



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from prestans import exception
from prestans.types import DataCollection
from prestans.types import DataStructure
from prestans.types import DataType

from prestans.types import Boolean
from prestans.types import Float
from prestans.types import Integer
from prestans.types import String

from prestans.types import Date
from prestans.types import DateTime
from prestans.types import Time


class Array(DataCollection):

    def __init__(self, required=True,element_template=None,
                 min_length=None, max_length=None, description=None):
        """
        :param required:
        :type required: bool
        :param element_template:
        :param min_length:
        :type min_length: int
        :param max_length:
        :type max_length: int
        :param description:
        :type description: str
        """

        if not isinstance(element_template, DataType):
            msg = "Array element_template must a DataType subclass; %s given" % element_template.__class__.__name__
            raise TypeError(msg)

        self._element_template = element_template  # type: DataType

        # force required to be True if basic type in  use
        if isinstance(element_template, DataType) and \
           not isinstance(element_template, DataCollection) and \
           not isinstance(element_template, DataStructure):
            element_template._required = True

        self._required = required
        self._min_length = min_length
        self._max_length = max_length
        self._description = description

        self._array_elements = list()

    def __len__(self):
        return len(self._array_elements)

    def __iter__(self):
        """"
        With a little help from:
        http://johnmc.co/llum/the-easiest-way-to-implement-__iter__-for-a-python-object
        """
        for element in self._array_elements:
            yield element

    def __getitem__(self, index):
        return self._array_elements[index]

    def __contains__(self, item):
        return item in self._array_elements

    @property
    def max_length(self):
        return self._max_length

    @property
    def min_length(self):
        return self._min_length

    @property
    def description(self):
        return self._description

    @property
    def is_scalar(self):
        """
        :return:
        :rtype: bool
        """
        return \
            isinstance(self._element_template, Boolean) or \
            isinstance(self._element_template, Float) or  \
            isinstance(self._element_template, Integer) or \
            isinstance(self._element_template, String)

    @property
    def element_template(self):
        return self._element_template

    @element_template.setter
    def element_template(self, value):
        self._element_template = value

    def blueprint(self):

        blueprint = dict()
        blueprint['type'] = 'array'

        constraints = dict()
        constraints['required'] = self._required
        constraints['min_length'] = self._min_length
        constraints['max_length'] = self._max_length
        constraints['element_template'] = self._element_template.blueprint()
        constraints['description'] = self._description

        blueprint['constraints'] = constraints
        return blueprint

    def remove(self, value):
        self._array_elements.remove(value)

    def validate(self, value, attribute_filter=None, minified=False):
        """
        :param value:
        :type value: list | None
        :param attribute_filter:
        :type attribute_filter: prestans.parser.AttributeFilter
        :param minified:
        :type minified: bool
        :return:
        """

        if not self._required and value is None:
            return None
        elif self._required and value is None:
            raise exception.RequiredAttributeError()

        _validated_value = self.__class__(
            element_template=self._element_template,
            min_length=self._min_length,
            max_length=self._max_length
        )

        if not isinstance(value, (list, tuple)):
            raise TypeError(value)

        for array_element in value:

            if isinstance(self._element_template, DataCollection):
                validated_array_element = self._element_template.validate(array_element, attribute_filter, minified)
            else:
                validated_array_element = self._element_template.validate(array_element)

            _validated_value.append(validated_array_element)

        if self._min_length is not None and len(_validated_value) < self._min_length:
            raise exception.LessThanMinimumError(value, self._min_length)

        if self._max_length is not None and len(_validated_value) > self._max_length:
            raise exception.MoreThanMaximumError(value, self._max_length)

        return _validated_value

    def append(self, value):

        if isinstance(value, (list, tuple)):

            for element in value:
                self.append(element)
            return

        # check for basic types supported by array
        if isinstance(self._element_template, Boolean) or \
           isinstance(self._element_template, Float) or \
           isinstance(self._element_template, Integer) or \
           isinstance(self._element_template, String):
            value = self._element_template.__class__().validate(value)
        elif isinstance(self._element_template, Date) or \
             isinstance(self._element_template, DateTime) or \
             isinstance(self._element_template, Time):
            value = self._element_template.__class__().validate(value)
        elif not isinstance(value, self._element_template.__class__):
            msg = "prestans array elements must be of type %s; given %s" % (
                self._element_template.__class__.__name__, value.__class__.__name__
            )
            raise TypeError(msg)

        self._array_elements.append(value)

    def as_serializable(self, attribute_filter=None, minified=False):

        _result_array = list()

        # convert filter to immutable if it isn't already
        from prestans.parser import AttributeFilter
        if isinstance(attribute_filter, AttributeFilter):
            attribute_filter = attribute_filter.as_immutable()

        for array_element in self._array_elements:

            if isinstance(self._element_template, DataCollection):
                serialized_value = array_element.as_serializable(attribute_filter, minified)
                _result_array.append(serialized_value)
            elif isinstance(self._element_template, DataStructure):
                serialized_value = self._element_template.as_serializable(array_element)
                _result_array.append(serialized_value)
            elif isinstance(self._element_template, DataType):
                _result_array.append(array_element)

        return _result_array

    def attribute_rewrite_map(self):
        if isinstance(self._element_template, DataCollection):
            return self._element_template.attribute_rewrite_map()
        else:
            return None

    def attribute_rewrite_reverse_map(self):
        if isinstance(self._element_template, DataCollection):
            return self._element_template.attribute_rewrite_reverse_map()
        else:
            return None

    def get_attribute_filter(self, default_value=False):

        attribute_filter = None

        if isinstance(self._element_template, DataCollection):
            attribute_filter = self._element_template.get_attribute_filter(default_value)
        elif isinstance(self._element_template, DataType) or isinstance(self._element_template, DataStructure):
            attribute_filter = default_value

        return attribute_filter



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import inspect

from prestans import exception
from prestans import parser
from prestans import types


class ModelAdapter(object):
    
    def __init__(self, rest_model_class, persistent_model_class):
        """
        :param rest_model_class:
        :param persistent_model_class:
        """

        if issubclass(rest_model_class, types.Model):
            self._rest_model_class = rest_model_class
        else:
            raise TypeError("rest_model_class must be sub class of prestans.types.Model")

        self._persistent_model_class = persistent_model_class

    @property
    def persistent_model_class(self):
        return self._persistent_model_class
    
    @property
    def rest_model_class(self):
        return self._rest_model_class

    def adapt_persistent_to_rest(self, persistent_object, attribute_filter=None):
        """
        adapts a persistent model to a rest model by inspecting
        """
        # convert filter to immutable if it isn't already
        if isinstance(attribute_filter, parser.AttributeFilter):
            attribute_filter = attribute_filter.as_immutable()

        rest_model_instance = self.rest_model_class()

        for attribute_key in rest_model_instance.get_attribute_keys():

            # attribute is not visible don't bother processing
            if isinstance(attribute_filter, (parser.AttributeFilter, parser.AttributeFilterImmutable)) and \
               not attribute_filter.is_attribute_visible(attribute_key):
                continue

            rest_attr = getattr(self.rest_model_class, attribute_key)

            # don't bother processing if the persistent model doesn't have this attribute
            if not hasattr(persistent_object, attribute_key):

                if isinstance(rest_attr, types.Model):
                    #: If the attribute is a Model, then we set it to None otherwise we get a model
                    #: with default values, which is invalid when constructing responses
                    try:
                        setattr(rest_model_instance, attribute_key, None)
                    # catch any exception thrown from setattr to give a usable error message
                    except TypeError as exp:
                        raise TypeError('Attribute %s, %s' % (attribute_key, str(exp)))

                continue
            # ignore class methods
            elif inspect.ismethod(getattr(persistent_object, attribute_key)):
                import logging
                logging.error("ignoring method: "+attribute_key)
                continue

            # handles prestans array population from SQLAlchemy relationships
            elif isinstance(rest_attr, types.Array):

                persistent_attr_value = getattr(persistent_object, attribute_key)
                rest_model_array_handle = getattr(rest_model_instance, attribute_key)

                # iterator uses the .append method exposed by prestans arrays to validate
                # and populate the collection in the instance.
                for collection_element in persistent_attr_value:
                    if rest_attr.is_scalar:
                        rest_model_array_handle.append(collection_element)
                    else:
                        element_adapter = registry.get_adapter_for_rest_model(rest_attr.element_template)

                        # check if there is a sub model filter
                        sub_attribute_filter = None
                        if attribute_filter and attribute_key in attribute_filter:
                            sub_attribute_filter = getattr(attribute_filter, attribute_key)

                        adapted_rest_model = element_adapter.adapt_persistent_to_rest(
                            collection_element,
                            sub_attribute_filter
                        )
                        rest_model_array_handle.append(adapted_rest_model)

            elif isinstance(rest_attr, types.Model):

                try:
                    persistent_attr_value = getattr(persistent_object, attribute_key)

                    if persistent_attr_value is None:
                        adapted_rest_model = None
                    else:
                        model_adapter = registry.get_adapter_for_rest_model(rest_attr)

                        # check if there is a sub model filter
                        sub_attribute_filter = None
                        if isinstance(attribute_filter, (parser.AttributeFilter, parser.AttributeFilterImmutable)) and \
                                attribute_key in attribute_filter:
                            sub_attribute_filter = getattr(attribute_filter, attribute_key)

                        adapted_rest_model = model_adapter.adapt_persistent_to_rest(
                            persistent_attr_value,
                            sub_attribute_filter
                        )

                    setattr(rest_model_instance, attribute_key, adapted_rest_model)

                except TypeError as exp:
                    raise TypeError('Attribute %s, %s' % (attribute_key, str(exp)))
                except exception.DataValidationException as exp:
                    raise exception.InconsistentPersistentDataError(attribute_key, str(exp))

            else:

                # otherwise copy the value to the rest model
                try:
                    persistent_attr_value = getattr(persistent_object, attribute_key)
                    setattr(rest_model_instance, attribute_key, persistent_attr_value)
                except TypeError as exp:
                    raise TypeError('Attribute %s, %s' % (attribute_key, str(exp)))
                except exception.ValidationError as exp:
                    raise exception.InconsistentPersistentDataError(attribute_key, str(exp))

        return rest_model_instance


def adapt_persistent_instance(persistent_object, target_rest_class=None, attribute_filter=None):
    """
    Adapts a single persistent instance to a REST model; at present this is a
    common method for all persistent backends.

    Refer to: https://groups.google.com/forum/#!topic/prestans-discuss/dO1yx8f60as
    for discussion on this feature
    """

    # try and get the adapter and the REST class for the persistent object
    if target_rest_class is None:
        adapter_instance = registry.get_adapter_for_persistent_model(persistent_object)
    else:
        if inspect.isclass(target_rest_class):
            target_rest_class = target_rest_class()

        adapter_instance = registry.get_adapter_for_persistent_model(persistent_object, target_rest_class)

    # would raise an exception if the attribute_filter differs from the target_rest_class
    if attribute_filter is not None and isinstance(attribute_filter, parser.AttributeFilter):
        parser.AttributeFilter.from_model(target_rest_class).conforms_to_template_filter(attribute_filter)

    # convert filter to immutable if it isn't already
    if isinstance(attribute_filter, parser.AttributeFilter):
        attribute_filter = attribute_filter.as_immutable()

    return adapter_instance.adapt_persistent_to_rest(persistent_object, attribute_filter)


def adapt_persistent_collection(persistent_collection, target_rest_class=None, attribute_filter=None):
    # ensure that collection is iterable and has at least one element
    persistent_collection_length = 0

    # attempt to detect the length of the persistent_collection
    if persistent_collection and isinstance(persistent_collection, (list, tuple)):
        persistent_collection_length = len(persistent_collection)
    # SQLAlchemy query
    elif persistent_collection and persistent_collection.__module__ == "sqlalchemy.orm.query":
        persistent_collection_length = persistent_collection.count()
    # Google App Engine NDB
    elif persistent_collection and persistent_collection.__module__ == "google.appengine.ext.ndb":
        persistent_collection_length = persistent_collection.count()

    # if the persistent_collection is empty then return a blank array
    if persistent_collection_length == 0:
        return types.Array(element_template=target_rest_class())

    # try and get the adapter and the REST class for the persistent object
    if target_rest_class is None:
        adapter_instance = registry.get_adapter_for_persistent_model(
            persistent_collection[0]
        )
    else:
        if inspect.isclass(target_rest_class):
            target_rest_class = target_rest_class()

        adapter_instance = registry.get_adapter_for_persistent_model(
            persistent_collection[0],
            target_rest_class
        )

    # would raise an exception if the attribute_filter differs from the target_rest_class
    if attribute_filter is not None and isinstance(attribute_filter, parser.AttributeFilter):
        parser.AttributeFilter.from_model(target_rest_class).conforms_to_template_filter(attribute_filter)

    # convert filter to immutable if it isn't already
    if isinstance(attribute_filter, parser.AttributeFilter):
        attribute_filter = attribute_filter.as_immutable()

    adapted_models = types.Array(element_template=adapter_instance.rest_model_class())

    for persistent_object in persistent_collection:
        adapted_models.append(adapter_instance.adapt_persistent_to_rest(persistent_object, attribute_filter))

    return adapted_models


class AdapterRegistryManager(object):
    """
    AdapterRegistryManager keeps track of rest to persistent model maps

    New AdapterRegistryManager's should not be instantiated by the application, a singleton
    instance is supplied by this package.
    """
    DEFAULT_REST_ADAPTER = "prestans_rest_default_adapter"

    def __init__(self):
        self._persistent_map = dict()
        self._rest_map = dict()

    @classmethod
    def generate_signature(cls, class_or_instance):
        if inspect.isclass(class_or_instance):
            return class_or_instance.__module__ + "." + class_or_instance.__name__
        else:
            return class_or_instance.__class__.__module__ + "." + class_or_instance.__class__.__name__

    def register_adapter(self, model_adapter):
        
        if not isinstance(model_adapter, ModelAdapter):
            msg = "Registry received instance of type %s is not a ModelAdapter" % model_adapter.__class__.__name__
            raise TypeError(msg)
        
        rest_class_signature = self.generate_signature(model_adapter.rest_model_class)
        persistent_class_signature = self.generate_signature(model_adapter.persistent_model_class)

        # store references to rest model
        self._rest_map[rest_class_signature] = model_adapter

        if persistent_class_signature not in self._persistent_map:
            self._persistent_map[persistent_class_signature] = dict()

        # store a reference to the adapter under both REST signature and default key
        # the default is always the last registered model (to match behaviour before this was patched)
        self._persistent_map[persistent_class_signature][self.DEFAULT_REST_ADAPTER] = model_adapter
        self._persistent_map[persistent_class_signature][rest_class_signature] = model_adapter

    def register_persistent_rest_pair(self, persistent_model_class, rest_model_class):
        """
        :param persistent_model_class:
        :param rest_model_class:
        """
        self.register_adapter(ModelAdapter(
            rest_model_class=rest_model_class,
            persistent_model_class=persistent_model_class
        ))

    def clear_registered_adapters(self):
        """
        Clears all of the currently registered model adapters
        """
        self._persistent_map.clear()
        self._rest_map.clear()

    def get_adapter_for_persistent_model(self, persistent_model, rest_model=None):
        """
        :param persistent_model: instance of persistent model
        :param rest_model: specific REST model
        :return: the matching model adapter
        :rtype: ModelAdapter
        """
        persistent_signature = self.generate_signature(persistent_model)
        
        if persistent_signature in self._persistent_map:
            sub_map = self._persistent_map[persistent_signature]

            # return the first match if REST model was not specified
            if rest_model is None:
                return self._persistent_map[persistent_signature][self.DEFAULT_REST_ADAPTER]
            else:
                rest_sig = self.generate_signature(rest_model)
                if rest_sig in sub_map:
                    return self._persistent_map[persistent_signature][rest_sig]

        raise TypeError("No registered Data Adapter for class %s" % persistent_signature)
        
    def get_adapter_for_rest_model(self, rest_model):
        """
        :param rest_model: instance of REST model
        :return: the matching model adapter
        :rtype: ModelAdapter
        """
        class_signature = self.generate_signature(rest_model)
        
        if class_signature not in self._rest_map:
            raise TypeError("No registered Data Adapter for class %s" % class_signature)

        return self._rest_map[class_signature]


# singleton instantiated if adapter package is imported
registry = AdapterRegistryManager()



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import inspect
import os
import sys

import prestans.devel.gen.closure
import prestans.types


class Inspector(object):

    def __init__(self, models_definition):
        self._models_definition = models_definition

        self._is_file = os.path.isfile(self._models_definition)

    def inspect(self):

        blueprints = list()

        if self._is_file:
            try:
                sys.path.append(os.path.dirname(self._models_definition))
                model_python_file = os.path.split(self._models_definition)[1:][0]
                namespace = model_python_file.split(".")[0]
                models = __import__(namespace, globals(), locals(), [])
            except ImportError as import_error:
                print ("\nFailed to process %s, %s" % (self._models_definition, import_error))
                sys.exit(1)
        else:
            import importlib
            models = importlib.import_module(name=self._models_definition)

        for name, type_instance in iter(models.__dict__.items()):
            if name.startswith('__') or inspect.ismethod(type_instance) or not inspect.isclass(type_instance) or\
             not issubclass(type_instance, prestans.types.Model):
                continue

            blueprint = type_instance().blueprint()
            for field_name, field_blueprint in iter(blueprint['fields'].items()):
                field_blueprint['map_name'] = type_instance().attribute_rewrite_map()[field_name]
            blueprints.append(blueprint)

        return blueprints


class Preplate(object):

    def __init__(self, template_type, models_definition, namespace, filter_namespace, output_directory):
        
        self._template_type = template_type
        self._models_definition = models_definition
        self._namespace = namespace
        self._filter_namespace = filter_namespace
        self._output_directory = output_directory

        from jinja2 import Environment
        from jinja2 import PackageLoader

        loader = PackageLoader('prestans', 'devel/gen/templates')
        self._template_engine = Environment(trim_blocks=True, loader=loader)
        self._template_engine.globals["isinstance"] = isinstance
        self._template_engine.globals["list"] = list

    @property
    def template_type(self):
        return self._template_type

    @property
    def models_definition(self):
        return self._models_definition

    @property
    def namespace(self):
        return self._namespace

    @property
    def filter_namespace(self):
        return self._filter_namespace

    @property
    def output_directory(self):
        return self._output_directory

    @property
    def template_engine(self):
        return self._template_engine

    def run(self):

        template = None
        if self._template_type == "closure.model":
            from prestans.devel.gen.closure import Model
            template = Model(
                template_engine=self._template_engine,
                models_definition=self.models_definition,
                namespace=self._namespace,
                filter_namespace=self._filter_namespace,
                output_directory=self._output_directory
            )
        elif self._template_type == "closure.filter":
            from prestans.devel.gen.closure import Filter
            template = Filter(
                template_engine=self._template_engine,
                models_definition=self.models_definition,
                namespace=self._namespace,
                output_directory=self._output_directory
            )

        if template is None:
            return 1
        
        return template.run()



from prestans.rest import Response


class DictionaryResponse(Response):
    """
    DictionaryResponse serializes dictionaries using the selected_serializer
    """

    @property
    def body(self):
        """
        Overridden response does not support md5, text or json properties. _app_iter
        is set using rules defined by prestans.

        body getter will return the validated prestans model.

        webob does the heavy lifting with headers.
        """
        return self._app_iter

    @body.setter
    def body(self, value):

        # value should be a dict
        if not isinstance(value, dict):
            raise TypeError("%s is not a dictionary" % value.__class__.__name__)

        #: _app_iter assigned to value
        #: we need to serialize the contents before we know the length
        #: deffer the content_length property to be set by getter
        self._app_iter = value

    def __call__(self, environ, start_response):

        if not isinstance(self.body, dict):
            raise TypeError("body is not a dict")

        start_response(self.status, self.headerlist)

        # attempt serializing via registered serializer
        body_as_string = self._selected_serializer.dumps(self.body)

        if not isinstance(body_as_string, str):
            raise TypeError("%s dumps must return a python str not %s" % (
                self._selected_serializer.__class__.__name__,
                body_as_string.__class__.__name__)
            )

        # set content_length
        self.content_length = len(body_as_string)

        return [body_as_string]



import webob

from prestans import deserializer
from prestans import exception
from prestans.http import VERB
from prestans.parser import AttributeFilter
from prestans.types import DataCollection


class Request(webob.Request):
    """
    Request is parsed REST Request; it's inherits and relies on Webob.Request to
    do the heavy lifting of parsing HTTP requests. It adds on top parsing of
    REST bodies and parameter sets based on rules set by the prestans app.

    It's responsible for making sense of the prestans headers and making them
    available to the RequestHandler
    """

    def __init__(self, environ, charset, logger, deserializers, default_deserializer):

        super(Request, self).__init__(environ=environ, charset=charset)
        self._logger = logger
        self._deserializers = deserializers
        self._default_deserializer = default_deserializer
        self._attribute_filter = None
        self._selected_deserializer = None
        self._parameter_set = None

        self._charset = charset

        self._body_template = None
        self._parsed_body = None

    @property
    def method(self):
        return self.environ['REQUEST_METHOD']

    @property
    def logger(self):
        return self._logger

    @property
    def parsed_body(self):

        if self.body_template is None:
            msg = "access to request.parsed_body is not allowed when body_template is set to None"
            raise AttributeError(msg)

        self.parse_body()

        return self._parsed_body

    @property
    def supported_mime_types(self):
        return [deserializer.content_type() for deserializer in self._deserializers]

    @property
    def supported_mime_types_str(self):
        return ''.join(str(mime_type) + ',' for mime_type in self.supported_mime_types)[:-1]

    @property
    def selected_deserializer(self):
        return self._selected_deserializer

    @property
    def default_deserializer(self):
        return self._default_deserializer

    def set_deserializer_by_mime_type(self, mime_type):
        """
        :param mime_type:
        :return:

        Used by content_type_set to set get a reference to the serializer object
        """

        for deserializer in self._deserializers:
            if deserializer.content_type() == mime_type:
                self._selected_deserializer = deserializer
                return

        raise exception.UnsupportedContentTypeError(mime_type, self.supported_mime_types_str)

    @property
    def attribute_filter(self):
        return self._attribute_filter

    @attribute_filter.setter
    def attribute_filter(self, value):
        """
        Attribute filter that used to parse the request_body. This must be set
        before the body_template parameter is set
        """
        self._attribute_filter = value

    @property
    def parameter_set(self):
        """
        Accessor for parameter_set
        """
        return self._parameter_set

    @parameter_set.setter
    def parameter_set(self, value):
        """
        Mutator for parameter_set
        """
        self._parameter_set = value

    @property
    def body_template(self):
        return self._body_template

    @body_template.setter
    def body_template(self, value):
        """
        Must be an instance of a prestans.types.DataCollection subclass; this is
        generally set during the RequestHandler lifecycle. Setting this spwans the
        parsing process of the body. If the HTTP verb is GET an AssertionError is
        thrown. Use with extreme caution.
        """

        if self.method == VERB.GET:
            raise AssertionError("body_template cannot be set for GET requests")

        if value is None:
            self.logger.warning("body_template is None, parsing will be ignored")
            return

        if not isinstance(value, DataCollection):
            msg = "body_template must be an instance of %s.%s" % (
                DataCollection.__module__,
                DataCollection.__name__
            )
            raise AssertionError(msg)

        self._body_template = value

        # get a deserializer based on the Content-Type header
        # do this here so the handler gets a chance to setup extra serializers
        self.set_deserializer_by_mime_type(self.content_type)

    def parse_body(self):

        if self._parsed_body is None and self._body_template is not None:
            # parse the body using the deserializer
            unserialized_body = self.selected_deserializer.loads(self.body)

            # validate the body using the template and attribute_filter
            self._parsed_body = self._body_template.validate(
                unserialized_body,
                self.attribute_filter,
                self.is_minified
            )

    def register_deserializers(self, deserializers):

        if not isinstance(deserializers, list):
            deserializers = [deserializers]

        # todo: should this prevent duplicates for same mime-type?

        for new_deserializer in deserializers:

            if not isinstance(new_deserializer, deserializer.Base):
                msg = "registered deserializer %s does not inherit from prestans.serializer.DeSerializer" % (
                      new_deserializer.__class__.__name__
                )
                raise TypeError(msg)

        self._deserializers = self._deserializers + deserializers

    def get_response_attribute_filter(self, template_filter, template_model=None):
        """
        Prestans-Response-Attribute-List can contain a client's requested
        definition for attributes required in the response. This should match
        the response_attribute_filter_template?

        :param template_filter:
        :param template_model: the expected model that this filter corresponds to
        :return:
        :rtype: None | AttributeFilter
        """

        if template_filter is None:
            return None

        if 'Prestans-Response-Attribute-List' not in self.headers:
            return None

        # header not set results in a None
        attribute_list_str = self.headers['Prestans-Response-Attribute-List']

        # deserialize the header contents
        json_deserializer = deserializer.JSON()
        attribute_list_dictionary = json_deserializer.loads(attribute_list_str)

        # construct an AttributeFilter
        attribute_filter = AttributeFilter(
            from_dictionary=attribute_list_dictionary,
            template_model=template_model
        )

        #: Check template? Do this even through we might have template_model
        #: in case users have made a custom filter
        evaluated_filter = attribute_filter.conforms_to_template_filter(template_filter)

        return evaluated_filter

    @property
    def is_minified(self):
        """
        :return:
        :rtype: bool
        """

        if 'Prestans-Minification' not in self.headers:
            return False

        return self.headers['Prestans-Minification'].upper() == "ON"



import sys

if sys.version_info < (3,):
    integer_types = (int, long)
else:
    integer_types = (int,)


if sys.version_info < (3,):
    string_types = (str, unicode)
else:
    string_types = (str,)



__all__ = [
    'Config',
    'VerbConfig',
    'AttributeFilter',
    'AttributeFilterImmutable',
    'ParameterSet'
]

from prestans.parser.attribute_filter import AttributeFilter
from prestans.parser.attribute_filter_immutable import AttributeFilterImmutable
from prestans.parser.config import Config
from prestans.parser.parameter_set import ParameterSet
from prestans.parser.verb_config import VerbConfig



from prestans.types import BinaryResponse
from prestans.types import DataCollection


class VerbConfig(object):
    """
    VerbConfig sets out rules for each HTTP Verb that your API will make available.
    These rules are used by prestans to validate requests and responses before
    handing over execution control to your handler.

    All verbs in use must provide at least a response_template which should be a
    subclass of prestans.types.DataCollection.
    """

    def __init__(self, response_template=None, response_attribute_filter_default_value=False,
                 parameter_sets=None, body_template=None, request_attribute_filter=None):

        """
        Each handler has a meta attribute called __verb_config__ this must be an instance
        of prestans.parser.Config which accepts six named parameters one for each supported
        HTTP verb (HEAD, GET, POST, PUT, DELETE, PATCH) each one of which must be an
        instance of prestans.parser.VerbConfig. A VerbConfig accepts the following named
        parameters (not all of them are supported across all HTTP verbs):

        * response_template an instance of a prestans.types.DataCollection subclass
          i.e a Model or an Array of Prestans DataType. This is what Prestans will use
          to validate the response your handler sends back to the client.
        * response_attribute_filter_default_value Prestans automatically creates an
          attribute filter based on the response_template by default Prestans exposes
          all it's attributes in the response, setting this to False will hide all
          attributes be default. Your handler code is responsible for toggling
          visibility in either instance.
        * parameter_sets an array of prestans.parser.ParameterSet instances
        * body_template an instance of a prestans.types.DataCollection subclass i.e
          a Model or an Array of Prestans DataType, this is what Prestans will use
          to validate the request sent to your handler. If validation of the incoming
          data fails, Prestans will not execute the associated verb in your handler.
        * request_attribute_filter is an attribute filter used to relax or tighten
          rules for the incoming data. This is particularly useful if you want
          to use portions of a model. Particularly useful for UPDATE requests.
        """
        from prestans.parser import AttributeFilter
        from prestans.parser import ParameterSet

        self._response_attribute_filter_template = None

        # response_template; required parameter
        if response_template is not None and \
                (not isinstance(response_template, DataCollection) and not isinstance(response_template, BinaryResponse)):
            raise TypeError("response_template of type %s must be an instance of \
                a prestans.types.DataCollection subclass" % response_template.__class__.__name__)

        if response_template is not None and isinstance(response_template, DataCollection):
            self.response_attribute_filter_template = AttributeFilter.from_model(
                model_instance=response_template,
                default_value=response_attribute_filter_default_value
            )
        else:
            self.response_attribute_filter_template = None

        self._response_template = response_template

        # turn parameter_sets single object into a list
        if parameter_sets is not None and not isinstance(parameter_sets, list):
            parameter_sets = [parameter_sets]

        # check that all parameter sets are of valid type
        if isinstance(parameter_sets, list):
            for parameter_set in parameter_sets:
                if not isinstance(parameter_set, ParameterSet):
                    raise TypeError("parameter_set of type %s must be an instance of \
                    prestans.parser.ParameterSet" % parameter_set.__class__.__name__)

            self._parameter_sets = parameter_sets
        else:
            self._parameter_sets = list()

        # body_template
        if body_template is not None and not \
                isinstance(body_template, DataCollection):
            raise TypeError(
                "body_template of type %s must be an instance of \
                a prestans.types.DataCollection subclass" % body_template.__class__.__name__)

        self._body_template = body_template

        # request_attribute_filter
        if request_attribute_filter is not None and \
           not isinstance(request_attribute_filter, AttributeFilter):
            raise TypeError("request_attribute_filter of type %s must an instance \
            of prestans.parser.AttributeFilter" % request_attribute_filter.__class__.__name__)

        self._request_attribute_filter = request_attribute_filter

    def blueprint(self):

        verb_config_blueprint = dict()

        if self._response_template is not None:
            verb_config_blueprint['response_template'] = self._response_template.blueprint()
        else:
            verb_config_blueprint['response_template'] = self._response_template

        verb_config_blueprint['parameter_sets'] = []
        for parameter_set in self._parameter_sets:
            if parameter_set is not None:
                verb_config_blueprint['parameter_sets'].append(parameter_set.blueprint())

        if self._body_template is not None:
            verb_config_blueprint['body_template'] = self._body_template.blueprint()
        else:
            verb_config_blueprint['body_template'] = self._body_template

        if self._request_attribute_filter is not None:
            verb_config_blueprint['request_attribute_filter'] = self._request_attribute_filter.blueprint()
        else:
            verb_config_blueprint['request_attribute_filter'] = self._request_attribute_filter

        return verb_config_blueprint

    @property
    def response_template(self):
        return self._response_template

    @property
    def response_attribute_filter_template(self):
        return self._response_attribute_filter_template

    @response_attribute_filter_template.setter
    def response_attribute_filter_template(self, value):
        self._response_attribute_filter_template = value

    @property
    def parameter_sets(self):
        return self._parameter_sets

    @property
    def body_template(self):
        return self._body_template

    @property
    def request_attribute_filter(self):
        return self._request_attribute_filter



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from prestans import exception
from prestans.types import DataType


class Boolean(DataType):

    def __init__(self, default=None, required=True, description=None):

        if default is None or \
           isinstance(default, bool):
            self._default = default
        else:
            raise TypeError("default must be of type bool or None")
        self._required = required
        self._description = description

    @property
    def required(self):
        return self._required

    @property
    def default(self):
        return self._default

    @property
    def description(self):
        return self._description

    def blueprint(self):

        blueprint = dict()
        blueprint['type'] = 'boolean'

        constraints = dict()
        constraints['default'] = self.default
        constraints['required'] = self.required
        constraints['description'] = self.description

        blueprint['constraints'] = constraints
        return blueprint

    def validate(self, value):

        if not self._required and self._default is None and value is None:
            return None
        elif self._required and self._default is None and value is None:
            raise exception.RequiredAttributeError()
        elif value is None and self.default is not None:
            value = self._default

        if not isinstance(value, bool):
            raise exception.ParseFailedError()

        return value



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import inspect

from prestans import exception
from prestans.types import Array, Date, DateTime, Float, Integer, String


class ParameterSet(object):
    """
    ParameterSet is a group of prestans.types that are expected as GET parameters

    ParameterSet defines rules and patterns in which they are acceptable.
    while ParserRuleSet is responsible for running the parse mechanism.
    """

    def getmembers(self):
        """
        :return: list of members as name, type tuples
        :rtype: list
        """
        return filter(
            lambda m: not m[0].startswith("__") and not inspect.isfunction(m[1]) and not inspect.ismethod(m[1]),
            inspect.getmembers(self.__class__)
        )

    def blueprint(self):
        """
        blueprint support, returns a partial dictionary
        """

        blueprint = dict()
        blueprint['type'] = "%s.%s" % (self.__module__, self.__class__.__name__)

        # Fields
        fields = dict()

        # inspects the attributes of a parameter set and tries to validate the input
        for attribute_name, type_instance in self.getmembers():

            # must be one of the following types
            if not isinstance(type_instance, String) and \
               not isinstance(type_instance, Float) and \
               not isinstance(type_instance, Integer) and \
               not isinstance(type_instance, Date) and \
               not isinstance(type_instance, DateTime) and \
               not isinstance(type_instance, Array):
                raise TypeError("%s should be instance of\
                 prestans.types.String/Integer/Float/Date/DateTime/Array" % attribute_name)

            if isinstance(type_instance, Array):
                if not isinstance(type_instance.element_template, String) and \
                   not isinstance(type_instance.element_template, Float) and \
                   not isinstance(type_instance.element_template, Integer):
                    raise TypeError("%s should be instance of \
                        prestans.types.String/Integer/Float/Array" % attribute_name)

            fields[attribute_name] = type_instance.blueprint()

        blueprint['fields'] = fields
        return blueprint

    def validate(self, request):
        """
        validate method for %ParameterSet

        Since the introduction of ResponseFieldListParser, the parameter _response_field_list
        will be ignored, this is a prestans reserved parameter, and cannot be used by apps.

        :param request: The request object to be validated
        :type request: webob.request.Request
        :return The validated parameter set
        :rtype: ParameterSet
        """

        validated_parameter_set = self.__class__()

        # Inspects the attributes of a parameter set and tries to validate the input
        for attribute_name, type_instance in self.getmembers():

            #: Must be one of the following types
            if not isinstance(type_instance, String) and \
               not isinstance(type_instance, Float) and \
               not isinstance(type_instance, Integer) and \
               not isinstance(type_instance, Date) and \
               not isinstance(type_instance, DateTime) and \
               not isinstance(type_instance, Array):
                raise TypeError("%s should be of type \
                    prestans.types.String/Integer/Float/Date/DateTime/Array" % attribute_name)

            if issubclass(type_instance.__class__, Array):

                if not isinstance(type_instance.element_template, String) and \
                   not isinstance(type_instance.element_template, Float) and \
                   not isinstance(type_instance.element_template, Integer):
                    raise TypeError("%s elements should be of \
                        type prestans.types.String/Integer/Float" % attribute_name)

            try:

                #: Get input from parameters
                #: Empty list returned if key is missing for getall
                if issubclass(type_instance.__class__, Array):
                    validation_input = request.params.getall(attribute_name)
                #: Key error thrown if key is missing for getone
                else:
                    try:
                        validation_input = request.params.getone(attribute_name)
                    except KeyError:
                        validation_input = None

                #: Validate input based on data type rules,
                #: raises DataTypeValidationException if validation fails
                validation_result = type_instance.validate(validation_input)

                setattr(validated_parameter_set, attribute_name, validation_result)

            except exception.DataValidationException as exp:
                raise exception.ValidationError(
                    message=str(exp),
                    attribute_name=attribute_name,
                    value=validation_input,
                    blueprint=type_instance.blueprint())

        return validated_parameter_set



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
__all__ = [
    'adapt_persistent_instance',
    'adapt_persistent_collection'
]

from prestans.ext.data import adapters


def adapt_persistent_instance(persistent_object, target_rest_class=None, attribute_filter=None):
    """
    Wrapper on adapters.adapt_persistent_instance for SQLAlchemy
    """
    return adapters.adapt_persistent_instance(persistent_object, target_rest_class, attribute_filter)


def adapt_persistent_collection(persistent_collection, target_rest_class=None, attribute_filter=None):
    """
    Wrapper on adapters.adapt_persistent_collection for SQLAlchemy
    """
    return adapters.adapt_persistent_collection(persistent_collection, target_rest_class, attribute_filter)
    

class ModelAdapter(adapters.ModelAdapter):

    def __init__(self, rest_model_class, persistent_model_class):
        import logging
        logger = logging.getLogger("prestans")
        logger.warning("direct use of %s has been deprecated please use %s instead" % (
            self.__module__ + "." + self.__class__.__name__,
            adapters.ModelAdapter.__module__ + "." + adapters.ModelAdapter.__name__
        ))
        super(ModelAdapter, self).__init__(rest_model_class, persistent_model_class)



import webob

from prestans import exception
from prestans.http import STATUS
from prestans.parser import AttributeFilter
from prestans import serializer
from prestans.types import Array
from prestans.types import BinaryResponse
from prestans.types import DataCollection
from prestans.types import Model


class Response(webob.Response):
    """
    Response is the writable HTTP response. It inherits and leverages
    from webob.Response to do the heavy lifting of HTTP Responses. It adds to
    webob.Response prestans customisations.

    Overrides content_type property to use prestans' serializers with the set body
    """

    def __init__(self, charset, logger, serializers, default_serializer):

        super(Response, self).__init__()

        self._logger = logger
        self._serializers = serializers
        self._default_serializer = default_serializer
        self._selected_serializer = None
        self._template = None
        self._app_iter = []
        self._minify = False
        self._attribute_filter = None
        self._template = None
        self._charset = charset

        #:
        #: IETF hash dropped the X- prefix for custom headers
        #: http://stackoverflow.com/q/3561381
        #: http://tools.ietf.org/html/draft-saintandre-xdash-00
        #:
        from prestans import __version__ as version
        if not isinstance(version, str):
            version = version.encode("latin1")

        self.headers.add('Prestans-Version', version)

    @property
    def minify(self):
        return self._minify

    @minify.setter
    def minify(self, value):
        self._minify = value

    @property
    def logger(self):
        return self._logger

    @property
    def supported_mime_types(self):
        return [serializer.content_type() for serializer in self._serializers]

    @property
    def supported_mime_types_str(self):
        return ''.join(str(mime_type) + ',' for mime_type in self.supported_mime_types)[:-1]

    @property
    def selected_serializer(self):
        return self._selected_serializer

    @property
    def default_serializer(self):
        return self._default_serializer

    def _set_serializer_by_mime_type(self, mime_type):
        """
        :param mime_type:
        :return:

        used by content_type_set to set get a reference to the appropriate serializer
        """

        # ignore if binary response
        if isinstance(self._app_iter, BinaryResponse):
            self.logger.info("ignoring setting serializer for binary response")
            return

        for available_serializer in self._serializers:
            if available_serializer.content_type() == mime_type:
                self._selected_serializer = available_serializer
                self.logger.info("set serializer for mime type: %s" % mime_type)
                return

        self.logger.info("could not find serializer for mime type: %s" % mime_type)
        raise exception.UnsupportedVocabularyError(mime_type, self.supported_mime_types_str)

    @property
    def template(self):
        """
        is an instance of prestans.types.DataType; mostly a subclass of prestans.types.Model
        """
        return self._template

    @template.setter
    def template(self, value):

        if value is not None and (not isinstance(value, DataCollection) and
                                  not isinstance(value, BinaryResponse)):
            raise TypeError("template in response must be of type prestans.types.DataCollection or subclass")

        self._template = value

    #:
    #: Attribute filter setup
    #:

    @property
    def attribute_filter(self):
        return self._attribute_filter

    @attribute_filter.setter
    def attribute_filter(self, value):

        if value is not None and not isinstance(value, AttributeFilter):
            msg = "attribute_filter in response must be of type prestans.types.AttributeFilter"
            raise TypeError(msg)

        self._attribute_filter = value

    def _content_type__get(self):
        """
        Get/set the Content-Type header (or None), *without* the
        charset or any parameters.

        If you include parameters (or ``;`` at all) when setting the
        content_type, any existing parameters will be deleted;
        otherwise they will be preserved.
        """
        header = self.headers.get('Content-Type')
        if not header:
            return None
        return header.split(';', 1)[0]

    def _content_type__set(self, value):

        # skip for responses that have no body
        if self.status_code in [STATUS.NO_CONTENT, STATUS.PERMANENT_REDIRECT, STATUS.TEMPORARY_REDIRECT]:
            self.logger.info("attempt to set Content-Type to %s being ignored due to empty response" % value)
            self._content_type__del()
        else:
            self._set_serializer_by_mime_type(value)

            if ';' not in value:
                header = self.headers.get('Content-Type', '')
                if ';' in header:
                    params = header.split(';', 1)[1]
                    value += ';' + params
            self.headers['Content-Type'] = value

            self.logger.info("Content-Type set to: %s" % value)

    def _content_type__del(self):
        self.headers.pop('Content-Type', None)

    # content_type; overrides webob.Response line 606
    content_type = property(
        _content_type__get,
        _content_type__set,
        _content_type__del,
        doc=_content_type__get.__doc__
    )

    # body; overrides webob.Response line 324
    @property
    def body(self):
        """
        Overridden response does not support md5, text or json properties. _app_iter
        is set using rules defined by prestans.

        body getter will return the validated prestans model.

        webob does the heavy lifting with headers.
        """

        #: If template is null; return an empty iterable
        if self.template is None:
            return []

        return self._app_iter

    @body.setter
    def body(self, value):

        #: If not response template; we have to assume its NO_CONTENT
        #: hence do not allow setting the body
        if self.template is None:
            raise AssertionError("response_template is None; handler can't return a response")

        #: value should be a subclass prestans.types.DataCollection
        if not isinstance(value, DataCollection) and \
                not isinstance(value, BinaryResponse):
            msg = "%s is not a prestans.types.DataCollection or prestans.types.BinaryResponse subclass" % (
                value.__class__.__name__
            )
            raise TypeError(msg)

        #: Ensure that it matches the return type template
        if not value.__class__ == self.template.__class__:
            msg = "body must of be type %s, given %s" % (
                self.template.__class__.__name__,
                value.__class__.__name__
            )
            raise TypeError(msg)

        #: If it's an array then ensure that element_template matches up
        if isinstance(self.template, Array) and \
           not isinstance(value.element_template, self.template.element_template.__class__):
            msg = "array elements must of be type %s, given %s" % (
                self.template.element_template.__class__.__name__,
                value.element_template.__class__.__name__
            )
            raise TypeError(msg)

        #: _app_iter assigned to value
        #: we need to serialize the contents before we know the length
        #: deffer the content_length property to be set by getter
        self._app_iter = value

    # body = property(_body__get, _body__set, _body__set)

    def register_serializers(self, serializers):
        """
        Adds extra serializers; generally registered during the handler lifecycle
        """
        for new_serializer in serializers:

            if not isinstance(new_serializer, serializer.Base):
                msg = "registered serializer %s.%s does not inherit from prestans.serializer.Serializer" % (
                    new_serializer.__module__,
                    new_serializer.__class__.__name__
                )
                raise TypeError(msg)

        self._serializers = self._serializers + serializers

    def __call__(self, environ, start_response):
        """
        Overridden WSGI application interface
        """

        # prestans equivalent of webob.Response line 1022
        if self.template is None or self.status_code == STATUS.NO_CONTENT:

            self.content_type = None

            start_response(self.status, self.headerlist)

            if self.template is not None:
                self.logger.warn("handler returns No Content but has a response_template; set template to None")

            return []

        # ensure what we are able to serialize is serializable
        if not isinstance(self._app_iter, DataCollection) and \
           not isinstance(self._app_iter, BinaryResponse):

            if isinstance(self._app_iter, list):
                app_iter_type = "list"
            else:
                app_iter_type = self._app_iter.__name__

            msg = "handler returns content of type %s; not a prestans.types.DataCollection subclass" % (
                app_iter_type
            )
            raise TypeError(msg)

        if isinstance(self._app_iter, DataCollection):

            #: See if attribute filter is completely invisible
            if self.attribute_filter is not None:

                #: Warning to say nothing is visible
                if not self.attribute_filter.are_any_attributes_visible():
                    self.logger.warn("attribute_filter has all the attributes turned \
                        off, handler will return an empty response")

                #: Warning to say none of the fields match
                model_attribute_filter = None
                if isinstance(self._app_iter, Array):
                    model_attribute_filter = AttributeFilter. \
                        from_model(self._app_iter.element_template)
                elif isinstance(self._app_iter, Model):
                    model_attribute_filter = AttributeFilter. \
                        from_model(self._app_iter)

                if model_attribute_filter is not None:
                    try:
                        model_attribute_filter.conforms_to_template_filter(self.attribute_filter)
                    except exception.AttributeFilterDiffers as exp:
                        exp.request = self.request
                        self.logger.warn("%s" % exp)

            # body should be of type DataCollection try; attempt calling
            # as_serializable with available attribute_filter
            serializable_body = self._app_iter.as_serializable(self.attribute_filter.as_immutable(), self.minify)

            #: attempt serializing via registered serializer
            stringified_body = self._selected_serializer.dumps(serializable_body)

            # if not isinstance(stringified_body, str):
            #     msg = "%s dumps must return a python str not %s" % (
            #         self._selected_serializer.__class__.__name__,
            #         stringified_body.__class__.__name__
            #     )
            #     raise TypeError(msg)

            #: set content_length
            self.content_length = len(stringified_body)

            start_response(self.status, self.headerlist)

            return [stringified_body.encode("utf-8")]

        elif isinstance(self._app_iter, BinaryResponse):

            if self._app_iter.content_length == 0 or \
                    self._app_iter.mime_type is None or \
                    self._app_iter.file_name is None:
                msg = "Failed to write binary response with content_length %i; mime_type %s; file_name %s" % (
                    self._app_iter.content_length,
                    self._app_iter.mime_type,
                    self._app_iter.file_name
                )
                self.logger.warn(msg)
                self.status = STATUS.INTERNAL_SERVER_ERROR
                self.content_type = "text/plain"
                return []

            # set the content type
            self.content_type = self._app_iter.mime_type

            #: Add content disposition header
            if self._app_iter.as_attachment:
                attachment = "attachment; filename=\"%s\"" % self._app_iter.file_name
                if not isinstance(attachment, str):
                    attachment = attachment.encode("latin1")

                self.headers.add("Content-Disposition", attachment)
            else:
                inline = "inline; filename=\"%s\"" % self._app_iter.file_name
                if not isinstance(inline, str):
                    inline = inline.encode("latin1")

                self.headers.add("Content-Disposition", inline)

            #: Write out response
            self.content_length = self._app_iter.content_length

            start_response(self.status, self.headerlist)
            return [self._app_iter.contents]

        else:
            raise AssertionError("prestans failed to write a binary or textual response")

    def __str__(self):
        #: Overridden so webob's __str__ skips serializing the body
        super(Response, self).__str__(skip_body=True)


import inspect

from prestans import exception
from prestans.http import STATUS
from prestans.http import VERB
from prestans import parser
from prestans import provider
from prestans.rest import ErrorResponse
from prestans import types


class RequestHandler(object):
    """
    RequestHandler is a callable that all API end-points must inherit from.
    end-points are instantiated by RequestRouter as a match for a URL.

    This class should not be initialised directly. Subclasses should
    override corresponding methods for HTTP verbs; get, post, delete, put, patch.
    """

    __provider_config__ = None
    __parser_config__ = None

    def __init__(self, args, kwargs, request, response, logger, debug):

        if self.__provider_config__ is None:
            self.__provider_config__ = provider.Config()
        if self.__parser_config__ is None:
            self.__parser_config__ = parser.Config()

        self._args = args
        self._kwargs = kwargs
        self._request = request
        self._response = response
        self._logger = logger
        self._debug = debug

        # initialization used elsewhere
        self.provider_authentication = None

    @property
    def request(self):
        return self._request

    @property
    def response(self):
        return self._response

    @property
    def logger(self):
        return self._logger

    @property
    def debug(self):
        return self._debug

    def blueprint(self):

        handler_blueprint = dict()

        signature_map = [
            VERB.GET,
            VERB.HEAD,
            VERB.POST,
            VERB.PUT,
            VERB.PATCH,
            VERB.DELETE,
            VERB.OPTIONS
        ]

        #: Provider configuration
        provider_blueprint = None

        if self.__class__.__provider_config__ is not None:
            provider_blueprint = self.__class__.__provider_config__.blueprint()

        handler_blueprint['provider_config'] = provider_blueprint

        # Make a list of methods supported by this handler
        for http_verb in signature_map:

            http_verb_lower = http_verb.lower()
            local_function_handle = getattr(self, http_verb_lower).__func__

            # Ignore if the local signature is the same as the base method
            if local_function_handle is getattr(RequestHandler, http_verb_lower).__func__:
                continue

            verb_blueprint = dict()

            # Docstring
            verb_blueprint['description'] = inspect.getdoc(local_function_handle)

            # Arguments, get the first set of parameters for the
            # function handle and ignore echoing self
            verb_blueprint['arguments'] = inspect.getargspec(local_function_handle)[0][1:]

            #: Parser configuration
            parser_blueprint = None

            if self.__class__.__parser_config__ is not None and \
                    self.__class__.__parser_config__.get_config_for_verb(http_verb) is not None:

                parser_config = self.__class__.__parser_config__.get_config_for_verb(http_verb)
                parser_blueprint = parser_config.blueprint()

            verb_blueprint['parser_config'] = parser_blueprint

            handler_blueprint[http_verb] = verb_blueprint

        return handler_blueprint

    def _setup_serializers(self):
        """
        Auto set the return serializer based on Accept headers
        http://docs.webob.org/en/latest/reference.html#header-getters

        Intersection of requested types and supported types tells us if we
        can in fact respond in one of the request formats
        """
        acceptable_offers = self.request.accept.acceptable_offers(self.response.supported_mime_types)
        if len(acceptable_offers) > 0:
            best_accept_match = acceptable_offers[0][0]
        else:
            best_accept_match = self.response.default_serializer.content_type()

        # best_accept_match = self.request.accept.best_match(
        #     self.response.supported_mime_types,
        #     default_match=self.response.default_serializer.content_type()
        # )

        self.logger.info("%s determined as best match for accept header: %s" % (
            best_accept_match,
            self.request.accept
        ))

        # if content_type is not acceptable it will raise UnsupportedVocabulary
        self.response.content_type = best_accept_match

    def __call__(self, environ, start_response):

        self.logger.info("handler %s.%s; callable execution start" % (self.__module__, self.__class__.__name__))

        try:
            # register additional serializers and de-serializers
            self.request.register_deserializers(self.register_deserializers())
            self.response.register_serializers(self.register_serializers())

            request_method = self.request.method

            # ensure we support the HTTP verb
            if not VERB.is_supported_verb(self.request.method):
                unimplemented_verb_error = exception.UnimplementedVerbError(self.request.method)
                unimplemented_verb_error.request = self.request
                raise unimplemented_verb_error

            #: Setup serializers
            self._setup_serializers()

            #: Authentication
            if self.__provider_config__.authentication is not None:
                self.__provider_config__.authentication.debug = self.debug
                self.provider_authentication = self.__provider_config__.authentication
                self.provider_authentication.request = self.request

            #: Configuration as provided by the API or default of a VerbConfig object
            verb_parser_config = self.__parser_config__.get_config_for_verb(request_method)

            #: Dress up the request and response with verb configuration
            if verb_parser_config is not None and verb_parser_config.response_template is not None:

                #: Set the response template and attribute filter
                self.response.template = verb_parser_config.response_template

                response_attr_filter_template = verb_parser_config. \
                    response_attribute_filter_template

                # minification support for response attribute filters
                rewrite_template_model = None
                if self.request.is_minified is True:

                    if isinstance(self.response.template, types.Array):
                        rewrite_template_model = self.response.template.element_template
                    else:
                        rewrite_template_model = self.response.template

                #: Response attribute filter
                self.response.attribute_filter = self.request.get_response_attribute_filter(
                    response_attr_filter_template,
                    rewrite_template_model
                )

                #: If the header is omitted then we ensure the response has a default template
                #: at this point we can assume that we are going to sent down a response
                if self.response.attribute_filter is None:
                    self.response.attribute_filter = response_attr_filter_template

            #: Parameter sets
            if verb_parser_config is not None and len(verb_parser_config.parameter_sets) > 0:

                for parameter_set in verb_parser_config.parameter_sets:

                    if not isinstance(parameter_set, parser.ParameterSet):
                        raise TypeError("%s not a subclass of ParameterSet" % parameter_set.__class__.__name__)

                    try:
                        validated_parameter_set = parameter_set.validate(self.request)

                        if validated_parameter_set is not None:
                            self.request.parameter_set = validated_parameter_set
                            break

                    except exception.DataValidationException as exp:
                        self.logger.error(exp)
                        error_response = ErrorResponse(exp, self.response.selected_serializer)
                        return error_response(environ, start_response)

            # parse body
            if not request_method == VERB.GET and verb_parser_config is not None:
                self.request.attribute_filter = verb_parser_config.request_attribute_filter
                #: Setting this runs the parser for the body
                #: Request will determine which serializer to use based on Content-Type
                self.request.body_template = verb_parser_config.body_template

            #: Warm up
            self.handler_will_run()

            try:
                #:
                #: See if the handler supports the called method
                #: prestans sets a sensible HTTP status code
                #:
                if request_method == VERB.GET:
                    self.get(*self._args, **self._kwargs)
                elif request_method == VERB.HEAD:
                    self.head(*self._args, **self._kwargs)
                elif request_method == VERB.POST:
                    self.post(*self._args, **self._kwargs)
                elif request_method == VERB.PUT:
                    self.put(*self._args, **self._kwargs)
                elif request_method == VERB.PATCH:
                    self.patch(*self._args, **self._kwargs)
                elif request_method == VERB.DELETE:
                    self.delete(*self._args, **self._kwargs)
                elif request_method == VERB.OPTIONS:
                    self.options(*self._args, **self._kwargs)
            except (exception.PermanentRedirect, exception.TemporaryRedirect) as exp:
                self._redirect(exp.url, exp.http_status)
            # re-raise all prestans exceptions
            except exception.Base as exp:
                if isinstance(exception, exception.HandlerException):
                    exp.request = self.request

                raise exp
            # handle any non-prestans exceptions
            except Exception as exp:
                self.handler_raised_exception(exp)
            # always run the tear down method
            finally:
                self.handler_did_run()

            self.logger.info("handler %s.%s; callable execution ends" % (
                self.__module__,
                self.__class__.__name__
            ))

            return self.response(environ, start_response)

        except exception.UnimplementedVerbError as exp:
            self.logger.error(exp)
            error_response = ErrorResponse(exp, self.response.selected_serializer)
            return error_response(environ, start_response)

    def register_serializers(self):
        return []

    def register_deserializers(self):
        return []

    def handler_will_run(self):
        return None

    def handler_did_run(self):
        return None

    #:
    #: Default handler for a raised exception return service unavailable
    #:
    def handler_raised_exception(self, exp):
        if self.debug:
            raise exp
        else:
            self.logger.error("handler %s.%s; exception raised: %s" % (
                self.__module__,
                self.__class__,
                exp
            ))
            raise exception.ServiceUnavailable()

    #:
    #: Placeholder functions for HTTP Verb; implementing handlers must override these
    #: if not overridden prestans returns a Not Implemented error
    #:

    def get(self, *args, **kwargs):
        unimplemented_verb_error = exception.UnimplementedVerbError(VERB.GET)
        unimplemented_verb_error.request = self.request
        raise unimplemented_verb_error

    def head(self, *args, **kwargs):
        unimplemented_verb_error = exception.UnimplementedVerbError(VERB.HEAD)
        unimplemented_verb_error.request = self.request
        raise unimplemented_verb_error

    def post(self, *args, **kwargs):
        unimplemented_verb_error = exception.UnimplementedVerbError(VERB.POST)
        unimplemented_verb_error.request = self.request
        raise unimplemented_verb_error

    def put(self, *args, **kwargs):
        unimplemented_verb_error = exception.UnimplementedVerbError(VERB.PUT)
        unimplemented_verb_error.request = self.request
        raise unimplemented_verb_error

    def patch(self, *args, **kwargs):
        unimplemented_verb_error = exception.UnimplementedVerbError(VERB.PATCH)
        unimplemented_verb_error.request = self.request
        raise unimplemented_verb_error

    def delete(self, *args, **kwargs):
        unimplemented_verb_error = exception.UnimplementedVerbError(VERB.DELETE)
        unimplemented_verb_error.request = self.request
        raise unimplemented_verb_error

    def options(self, *args, **kwargs):
        unimplemented_verb_error = exception.UnimplementedVerbError(VERB.OPTIONS)
        unimplemented_verb_error.request = self.request
        raise unimplemented_verb_error

    def _redirect(self, url, status=STATUS.TEMPORARY_REDIRECT):
        self._response.status = status
        self._response.headers.add("Location", url)

    def redirect(self, url, status=STATUS.TEMPORARY_REDIRECT):

        self.logger.warn("direct use of %s.%s has been deprecated please raise an exception instead" % (
            self.__module__,
            "redirect"
        ))

        self._redirect(url, status)



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from prestans import exception
from prestans.types import DataType


class Float(DataType):

    def __init__(self, default=None, minimum=None, maximum=None, required=True,
                 choices=None, description=None):

        if minimum and maximum and minimum > maximum:
            raise ValueError("maximum cannot be less than minimum")

        self._default = default
        self._minimum = minimum
        self._maximum = maximum
        self._required = required
        self._choices = choices
        self._description = description

    @property
    def default(self):
        return self._default

    @property
    def minimum(self):
        return self._minimum

    @property
    def maximum(self):
        return self._maximum

    @property
    def required(self):
        return self._required

    @property
    def choices(self):
        return self._choices

    @property
    def description(self):
        return self._description

    def blueprint(self):

        blueprint = dict()
        blueprint['type'] = 'float'

        constraints = dict()
        constraints['default'] = self._default
        constraints['minimum'] = self._minimum
        constraints['maximum'] = self._maximum
        constraints['required'] = self._required
        constraints['choices'] = self._choices
        constraints['description'] = self._description

        blueprint['constraints'] = constraints
        return blueprint

    def validate(self, value):

        if not self._required and self._default is None and value is None:
            return None
        if self._required and self._default is None and value is None:
            raise exception.RequiredAttributeError()
        elif self._required and value is None:
            value = self._default
        elif not self._required and value is None:
            value = self._default

        try:
            _validated_value = float(value)
        except Exception as exp:
            raise exception.ParseFailedError("float encoding failed %s" % exp)

        if _validated_value is not None and self._minimum is not None and _validated_value < self._minimum:
            raise exception.LessThanMinimumError(value, self._minimum)

        if _validated_value is not None and self._maximum is not None and _validated_value > self._maximum:
            raise exception.MoreThanMaximumError(value, self._maximum)

        if self._choices is not None and _validated_value not in self._choices:
            raise exception.InvalidChoiceError(value, self._choices)

        return _validated_value



import inspect

from prestans.http import STATUS
from prestans.rest import RequestHandler


class BlueprintHandler(RequestHandler):
    """
    Generates a blueprint of the handler and serializes it using the requested Accept header.
    """

    def __init__(self, args, request, response, logger, debug, route_map):

        super(BlueprintHandler, self).__init__(args, request, response, logger, debug)
        self._route_map = route_map

    def _create_blueprint(self):

        blueprint_groups = dict()

        # interrogate each handler
        for regexp, handler_class in self.route_map:

            # Ignore discovery handler
            if issubclass(handler_class, BlueprintHandler):
                continue

            handler_blueprint = dict()
            handler_blueprint['url'] = regexp
            handler_blueprint['handler_class'] = handler_class.__name__
            handler_blueprint['description'] = inspect.getdoc(handler_class)
            handler_blueprint['supported_methods'] = handler_class(
                self._args,
                self.request,
                self.response,
                self.logger,
                self.debug
            ).blueprint()

            # Make a new group per module if one doesnt' exist
            if handler_class.__module__ not in blueprint_groups:
                blueprint_groups[handler_class.__module__] = []

            blueprint_groups[handler_class.__module__].append(handler_blueprint)

        return blueprint_groups

    @property
    def route_map(self):
        return self._route_map

    @route_map.setter
    def route_map(self, value):
        self._route_map = value

    def __call__(self, environ, start_response):

        # setup serializers
        self._setup_serializers()

        self.response.status = STATUS.OK

        self.response.body = self._create_blueprint()

        return self.response(environ, start_response)



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import base64

from prestans import exception
from prestans.types import DataStructure


class DataURLFile(DataStructure):
    """
    Accepts a Fileupload as part of the JSON body using FileReader's readAsDataURL

    readAsDataURL, encodes the contents of the file as a DataURLScheme,
    http://en.wikipedia.org/wiki/Data_URI_scheme

    Example
    http://www.html5rocks.com/en/tutorials/file/dndfiles/

    Meta information about the file upload is up to the implementing application
    """

    @classmethod
    def generate_filename(cls):
        import uuid
        return uuid.uuid4().hex

    def __init__(self, required=True, allowed_mime_types=None, description=None):

        if allowed_mime_types is None:
            allowed_mime_types = []
        elif isinstance(allowed_mime_types, str):
            allowed_mime_types = [allowed_mime_types]

        self._required = required
        self._allowed_mime_types = allowed_mime_types
        self._description = description

        self._mime_type = None
        self._file_contents = None

    @property
    def required(self):
        return self._required

    @property
    def allowed_mime_types(self):
        return self._allowed_mime_types

    @property
    def description(self):
        return self._description

    def blueprint(self):

        blueprint = dict()
        blueprint['type'] = 'data_url_file'

        constraints = dict()
        constraints['required'] = self._required
        constraints['allowed_mime_types'] = self._allowed_mime_types
        constraints['description'] = self._description

        blueprint['constraints'] = constraints
        return blueprint

    @property
    def mime_type(self):
        return self._mime_type

    @property
    def file_contents(self):
        return self._file_contents

    @property
    def base64_contents(self):
        return base64.b64encode(self._file_contents)

    def validate(self, value):

        _validated_value = self.__class__()

        if self._required and value is None:
            raise exception.RequiredAttributeError()

        if self._required is False and value is None:
            return value

        try:
            data_url, delimiter, base64_content = value.partition(',')
            _validated_value._mime_type = data_url.replace(';base64', '').replace('data:', '')
            _validated_value._file_contents = base64.b64decode(base64_content)
        except Exception as exp:
            raise exception.ParseFailedError("data url file encoding failed %s" % exp)

        if self._allowed_mime_types and len(self._allowed_mime_types) > 0 \
           and _validated_value._mime_type not in self._allowed_mime_types:
            raise exception.InvalidChoiceError(_validated_value._mime_type, self._allowed_mime_types)

        return _validated_value

    def save(self, path):
        """
        Writes file to a particular location

        This won't work for cloud environments like Google's App Engine, use with caution
        ensure to catch exceptions so you can provide informed feedback.

        prestans does not mask File IO exceptions so your handler can respond better.
        """

        file_handle = open(path, 'wb')
        file_handle.write(self._file_contents)
        file_handle.close()

    def as_serializable(self, value):
        #: This is passed in a DataURLFile and we construct a String back from it
        return "data:%s;base64,%s" % (value.mime_type, value.base64_contents.decode("utf-8"))



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import copy
import inspect
import string

from prestans import exception
from prestans.types import DataCollection
from prestans.types import DataStructure
from prestans.types import DataType


class Model(DataCollection):

    def __init__(self, required=True, description=None, **kwargs):
        """
        If you are using the Model constructor to provide Meta data, you can provide it
        a default dictionary to initialise instance to initialise it from

        Can run _create_instance_attributes to copy attribute templates to the instance,
        Model instances do not have any configurable options.

        @param required whether or not this model is required when used as an attribute
        @param default provides a default value for your type, used if one is not provided
        @param kwargs Named parameters of which to instantiate the model with
        """

        self._required = required
        self._description = description

        self._templates = {}
        self._attributes = {}
        self._create_instance_attributes(kwargs)

    def getmembers(self):
        """
        :return: list of members as name, type tuples
        :rtype: list
        """
        return filter(
            lambda m: not m[0].startswith("__") and not inspect.isfunction(m[1]) and not inspect.ismethod(m[1]),
            inspect.getmembers(self.__class__)
        )

    def attribute_count(self):

        attribute_count = 0

        for attribute_name, type_instance in self.getmembers():

            if isinstance(type_instance, DataType):
                attribute_count += 1

        return attribute_count

    def blueprint(self):

        blueprint = dict()
        blueprint['type'] = 'model'
        blueprint['description'] = inspect.getdoc(self.__class__)

        constraints = dict()
        constraints['required'] = self._required
        constraints['model_template'] = self.__class__.__name__
        constraints['description'] = self._description

        blueprint['constraints'] = constraints

        fields = dict()
        for attribute_name, type_instance in self.getmembers():

            if not isinstance(type_instance, DataType):
                raise TypeError("%s must be of a DataType subclass" % attribute_name)

            fields[attribute_name] = type_instance.blueprint()

        blueprint['fields'] = fields
        return blueprint

    def __getattribute__(self, key):
        value = object.__getattribute__(self, key)

        if key[0:1] == "_":
            return value

        attributes = self.__dict__["_attributes"]
        templates = self.__dict__["_templates"]

        if value is not None and key not in templates:
            return value

        validator = templates.get(key)
        # grab the value from local attribute dictionary for prestans types
        if validator is not None:
            value = attributes.get(key)

        # if attribute is a data collection and no value found we need to copy the template
        if isinstance(validator, DataCollection) and value is None:
            copied = copy.deepcopy(validator)

            self._attributes[key] = copied
            value = copied

        return value

    def __setattr__(self, key, value):

        if key[0:1] == "_":
            self.__dict__[key] = value
            return

        validator = self._templates.get(key)
        if validator is None:
            raise KeyError("No key named: %s in instance of type: %s" % (key, self.__class__.__name__))

        try:
            # if given an instance of data collection we can directly set it
            if isinstance(validator, DataCollection) and validator.__class__ == value.__class__:
                self._attributes[key] = value
            else:
                self._attributes[key] = validator.validate(value)

        except exception.DataValidationException as exp:
            raise exception.ValidationError(
                message=str(exp),
                attribute_name=key,
                value=value,
                blueprint=validator.blueprint()
            )

    def _create_instance_attributes(self, arguments):
        """
        Copies class level attribute templates and makes instance placeholders

        This step is required for direct uses of Model classes. This creates a
        copy of attribute_names ignores methods and private variables.
        DataCollection types are deep copied to ignore memory reference conflicts.

        DataType instances are initialized to None or default value.
        """
        for attribute_name, type_instance in self.getmembers():
            if isinstance(type_instance, DataType):
                self._templates[attribute_name] = type_instance

                value = None
                if attribute_name in arguments:
                    value = arguments[attribute_name]

                try:
                    self._attributes[attribute_name] = type_instance.validate(value)
                # we can safely ignore required warnings during initialization
                except exception.RequiredAttributeError:
                    self._attributes[attribute_name] = None

    def get_attribute_keys(self):
        """
        Returns a list of managed attributes for the Model class

        Implemented for use with data adapters, can be used to quickly make a list of the
        attribute names in a prestans model
        """

        _attribute_keys = list()

        for attribute_name, type_instance in self.getmembers():

            if isinstance(type_instance, DataType):
                _attribute_keys.append(attribute_name)

        return _attribute_keys

    def get_attribute_filter(self, default_value=False):
        from prestans.parser import AttributeFilter

        attribute_filter = AttributeFilter()

        for attribute_name, type_instance in self.getmembers():

            if isinstance(type_instance, DataCollection):
                setattr(attribute_filter, attribute_name, type_instance.get_attribute_filter(default_value))
            else:
                setattr(attribute_filter, attribute_name, default_value)

        return attribute_filter

    def validate(self, value, attribute_filter=None, minified=False):
        """
        :param value: serializable input to validate
        :type value: dict | None
        :param attribute_filter:
        :type: prestans.parser.AttributeFilter | None
        :param minified: whether or not the input is minified
        :type minified: bool
        :return: the validated model
        :rtype: Model
        """

        if self._required and (value is None or not isinstance(value, dict)):
            """
            Model level validation requires a parsed dictionary
            this is done by the serializer
            """
            raise exception.RequiredAttributeError()

        if not self._required and not value:
            """
            Value was not provided by caller, but require a template
            """
            return None

        _model_instance = self.__class__()

        rewrite_map = self.attribute_rewrite_map()

        from prestans.parser import AttributeFilter
        from prestans.parser import AttributeFilterImmutable

        for attribute_name, type_instance in self.getmembers():
            if not isinstance(type_instance, DataType):
                raise TypeError("%s must be a DataType subclass" % attribute_name)

            if isinstance(attribute_filter, (AttributeFilter, AttributeFilterImmutable)) and \
               not attribute_filter.is_attribute_visible(attribute_name):
                _model_instance._attributes[attribute_name] = None
                continue

            validation_input = None

            input_value_key = attribute_name

            # minification support
            if minified is True:
                input_value_key = rewrite_map[attribute_name]

            if input_value_key in value:
                validation_input = value[input_value_key]

            try:

                if isinstance(type_instance, DataCollection):
                    sub_attribute_filter = None
                    if attribute_filter and attribute_name in attribute_filter:
                        sub_attribute_filter = getattr(attribute_filter, attribute_name)

                    validated_object = type_instance.validate(
                        validation_input,
                        sub_attribute_filter,
                        minified
                    )
                else:
                    validated_object = type_instance.validate(validation_input)

                _model_instance._attributes[attribute_name] = validated_object

            except exception.DataValidationException as exp:
                raise exception.ValidationError(
                    message=str(exp),
                    attribute_name=attribute_name,
                    value=validation_input,
                    blueprint=type_instance.blueprint()
                )

        return _model_instance

    def attribute_rewrite_map(self):
        """
        Example: long_name -> a_b

        :return: the rewrite map
        :rtype: dict
        """

        rewrite_map = dict()
        token_rewrite_map = self.generate_attribute_token_rewrite_map()

        for attribute_name, type_instance in self.getmembers():

            if isinstance(type_instance, DataType):
                attribute_tokens = attribute_name.split('_')

                rewritten_attribute_name = ''
                for token in attribute_tokens:
                    rewritten_attribute_name += token_rewrite_map[token] + "_"
                # remove the trailing underscore
                rewritten_attribute_name = rewritten_attribute_name[:-1]

                rewrite_map[attribute_name] = rewritten_attribute_name

        return rewrite_map

    def attribute_rewrite_reverse_map(self):
        """
        Example: a_b -> long_name

        :return: the reverse rewrite map
        :rtype: dict
        """

        rewrite_map = dict()

        token_rewrite_map = self.generate_attribute_token_rewrite_map()

        for attribute_name, type_instance in self.getmembers():

            if isinstance(type_instance, DataType):

                attribute_tokens = attribute_name.split('_')
                rewritten_attribute_name = ''
                for token in attribute_tokens:
                    rewritten_attribute_name += token_rewrite_map[token] + "_"
                # remove the trailing underscore
                rewritten_attribute_name = rewritten_attribute_name[:-1]

                rewrite_map[rewritten_attribute_name] = attribute_name

        return rewrite_map

    def __contains__(self, attribute_name):

        has_key = attribute_name in self.__class__.__dict__

        if not has_key:

            base_classes = list()
            base_classes += list(self.__class__.__bases__)

            while not has_key and base_classes:

                # take a base class from the list
                base_class = base_classes.pop()

                # we found it clear the list
                if attribute_name in base_class.__dict__:
                    has_key = True
                    base_classes = list()
                # add any more base classes from this class
                else:
                    base_classes += list(base_class.__bases__)

        return has_key

    def generate_attribute_token_rewrite_map(self):

        rewrite_tokens = self.generate_attribute_tokens()
        minified_tokens = self.generate_minified_keys(len(rewrite_tokens))

        return dict(zip(rewrite_tokens, minified_tokens))

    def generate_attribute_tokens(self):

        rewrite_tokens = list()

        # create a list of tokens
        for attribute_name, type_instance in self.getmembers():

            if isinstance(type_instance, DataType):
                rewrite_tokens = rewrite_tokens + attribute_name.split('_')

        # remove duplicated; sort alphabetically for the algorithm to work
        rewrite_tokens = list(set(rewrite_tokens))
        rewrite_tokens.sort()

        return rewrite_tokens

    @classmethod
    def generate_minified_keys(cls, length=26, prefix=''):

        minified_keys = list()

        overflow = 0
        if length > 26:
            overflow = length - 26
            length = 26

        for index in range(0, length):

            generated_char = cls.generate_attribute_key(index)
            minified_keys.append(prefix + generated_char)

            if overflow > 0:

                sublist_length = overflow
                if sublist_length > 26:
                    sublist_length = 26

                sublist = cls.generate_minified_keys(sublist_length, generated_char)
                minified_keys = minified_keys + sublist
                overflow = overflow - len(sublist)

        if prefix == '':
            minified_keys.sort(key=len, reverse=False)

        return minified_keys

    @classmethod
    def generate_attribute_key(cls, val):
        """
        :param val:
        :type val: int
        :return:
        :rtype: str
        """
        return string.ascii_lowercase[val % 26] * (val // 26 + 1)

    def as_serializable(self, attribute_filter=None, minified=False):
        """
        Returns a dictionary with attributes and pure python representation of
        the data instances. If an attribute filter is provided as_serializable
        will respect the visibility.

        The response is used by serializers to return data to client

        :param attribute_filter:
        :type attribute_filter: prestans.parser.AttributeFilter
        :param minified:
        :type minified: bool
        """
        from prestans.parser import AttributeFilter
        from prestans.parser import AttributeFilterImmutable
        from prestans.types import Array

        model_dictionary = dict()

        rewrite_map = self.attribute_rewrite_map()

        # convert filter to immutable if it isn't already
        if isinstance(attribute_filter, AttributeFilter):
            attribute_filter = attribute_filter.as_immutable()

        for attribute_name, type_instance in self.getmembers():

            if isinstance(attribute_filter, (AttributeFilter, AttributeFilterImmutable)) and \
               not attribute_filter.is_attribute_visible(attribute_name):
                continue

            # support minification
            serialized_attribute_name = attribute_name
            if minified is True:
                serialized_attribute_name = rewrite_map[attribute_name]

            if attribute_name not in self._attributes or self._attributes[attribute_name] is None:
                if isinstance(type_instance, Array):
                    model_dictionary[serialized_attribute_name] = []
                else:
                    model_dictionary[serialized_attribute_name] = None
                continue

            if isinstance(type_instance, DataCollection):

                sub_attribute_filter = None
                if isinstance(attribute_filter, (AttributeFilter, AttributeFilterImmutable)) and attribute_name in attribute_filter:
                    sub_attribute_filter = getattr(attribute_filter, attribute_name)

                model_dictionary[serialized_attribute_name] = self._attributes[attribute_name].as_serializable(sub_attribute_filter, minified)

            elif isinstance(type_instance, DataStructure):
                python_value = self._attributes[attribute_name]
                serializable_value = type_instance.as_serializable(python_value)
                model_dictionary[serialized_attribute_name] = serializable_value

            elif isinstance(type_instance, DataType):
                model_dictionary[serialized_attribute_name] = self._attributes[attribute_name]

        return model_dictionary



import inspect

from prestans import exception
from prestans.types import Array
from prestans.types import DataCollection


class AttributeFilter(object):
    """
      {
        field_name0: true,
        field_name1: false,
        collection_name0: true,
        collection_name1: false,
        collection_name2: {
            sub_field_name0: true,
            sub_field_name1: false
        }
      }
    """

    def __init__(self, from_dictionary=None, template_model=None, is_array_scalar=False, **kwargs):
        """
        Creates an attribute filter object, optionally populates from a
        dictionary of booleans
        """
        # todo: is_array_scalar currently does nothing, fix or remove
        # todo: adding it as property breaks keys and other methods in current implementation

        if from_dictionary:
            self._init_from_dictionary(from_dictionary, template_model)

        for name, value in iter(kwargs.items()):
            if name in self.__dict__:
                setattr(self, name, value)
            else:
                raise KeyError(name)

    @classmethod
    def from_model(cls, model_instance, default_value=False, **kwargs):
        """
        wrapper for Model's get_attribute_filter
        """

        if not isinstance(model_instance, DataCollection):
            raise TypeError("model_instance must be a subclass of \
                prestans.types.DataCollection, %s given" % (model_instance.__class__.__name__))
        elif isinstance(model_instance, Array) and model_instance.is_scalar:
            return AttributeFilter(is_array_scalar=True)
        attribute_filter_instance = model_instance.get_attribute_filter(default_value)

        # kwargs support
        for name, value in iter(kwargs.items()):
            if name in attribute_filter_instance:
                setattr(attribute_filter_instance, name, value)
            else:
                raise KeyError(name)

        return attribute_filter_instance

    def blueprint(self):
        """
        :return: blueprint
        :rtype: dict
        """
        blueprint = dict()
        for key in self.keys():
            blueprint[key] = self.is_attribute_visible(key)

        return blueprint

    def conforms_to_template_filter(self, template_filter):
        """
        Check AttributeFilter conforms to the rules set by the template

         - If self, has attributes that template_filter does not contain, throw Exception
         - If sub list found, perform the first check
         - If self has a value for an attribute, assign to final AttributeFilter
         - If not found, assign value from template

         todo: rename as current name is mis-leading
        """

        if not isinstance(template_filter, self.__class__):
            raise TypeError("AttributeFilter can only check conformance against \
                another template filter, %s provided" % template_filter.__class__.__name__)

        # keys from the template
        template_filter_keys = template_filter.keys()
        # Keys from the object itself
        this_filter_keys = self.keys()

        # 1. Check to see if the client has provided unwanted keys
        unwanted_keys = set(this_filter_keys) - set(template_filter_keys)
        if len(unwanted_keys) > 0:
            raise exception.AttributeFilterDiffers(list(unwanted_keys))

        # 2. Make a attribute_filter that we send back
        evaluated_attribute_filter = AttributeFilter()

        # 3. Evaluate the differences between the two, with template_filter as the standard
        for template_key in template_filter_keys:

            if template_key in this_filter_keys:

                value = getattr(self, template_key)

                # if sub filter and boolean provided with of true, create default filter with value of true
                if isinstance(value, bool) and \
                   value is True and \
                   isinstance(getattr(template_filter, template_key), AttributeFilter):
                    setattr(evaluated_attribute_filter, template_key, getattr(template_filter, template_key))
                elif isinstance(value, bool):
                    setattr(evaluated_attribute_filter, template_key, value)
                elif isinstance(value, self.__class__):
                    # Attribute lists sort themselves out, to produce sub Attribute Filters
                    template_sub_list = getattr(template_filter, template_key)
                    this_sub_list = getattr(self, template_key)
                    setattr(
                        evaluated_attribute_filter, template_key,
                        this_sub_list.conforms_to_template_filter(template_sub_list)
                    )
            else:
                setattr(evaluated_attribute_filter, template_key, getattr(template_filter, template_key))

        return evaluated_attribute_filter

    def keys(self):
        """
        :returns: a list of usable keys
        :rtype: list

        """

        keys = list()

        for attribute_name, type_instance in inspect.getmembers(self):

            # ignore parameters with __ and if they are methods
            if attribute_name.startswith('__') or inspect.ismethod(type_instance):

                continue

            keys.append(attribute_name)

        return keys

    def __contains__(self, key):
        return key in self.__dict__

    def is_filter_at_key(self, key):
        """
        return True if attribute is a sub filter
        """

        if key in self:
            attribute_status = getattr(self, key)
            if isinstance(attribute_status, self.__class__):
                return True

        return False

    def is_attribute_visible(self, key):
        """
        Returns True if an attribute is visible
        If attribute is an instance of AttributeFilter, it returns True if all attributes
        of the sub filter are visible.

        :param key: name of attribute to check
        :type key: str
        :return: whether attribute is visible
        :rtype: bool
        """
        if key in self:
            attribute_status = getattr(self, key)
            if isinstance(attribute_status, bool) and attribute_status is True:
                return True
            elif isinstance(attribute_status, self.__class__) and attribute_status.are_any_attributes_visible():
                return True

        return False

    def are_any_attributes_visible(self):
        """
        checks to see if any attributes are set to true
        """

        for attribute_name, type_instance in inspect.getmembers(self):

            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                continue

            if isinstance(type_instance, bool) and type_instance is True:
                return True
            elif isinstance(type_instance, self.__class__) and type_instance.are_all_attributes_visible() is True:
                return True

        return False

    def are_all_attributes_visible(self):
        """
        checks to see if all attributes are set to true
        """

        for attribute_name, type_instance in inspect.getmembers(self):

            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                # Ignore parameters with __ and if they are methods
                continue

            if isinstance(type_instance, bool) and type_instance is False:
                return False
            elif isinstance(type_instance, self.__class__) and type_instance.are_all_attributes_visible() is False:
                return False

        return True

    def set_all_attribute_values(self, value):
        """
        sets all the attribute values to the value and propagate to any children
        """

        for attribute_name, type_instance in inspect.getmembers(self):

            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                # Ignore parameters with __ and if they are methods
                continue

            if isinstance(type_instance, bool):
                self.__dict__[attribute_name] = value
            elif isinstance(type_instance, self.__class__):
                type_instance.set_all_attribute_values(value)

    def as_dict(self):
        """
        turns attribute filter object into python dictionary
        """

        output_dictionary = dict()

        for attribute_name, type_instance in inspect.getmembers(self):

            if attribute_name.startswith('__') or inspect.ismethod(type_instance):
                continue

            if isinstance(type_instance, bool):
                output_dictionary[attribute_name] = type_instance
            elif isinstance(type_instance, self.__class__):
                output_dictionary[attribute_name] = type_instance.as_dict()

        return output_dictionary

    def _init_from_dictionary(self, from_dictionary, template_model=None):
        """
        Private helper to init values from a dictionary, wraps children into
        AttributeFilter objects

        :param from_dictionary: dictionary to get attribute names and visibility from
        :type from_dictionary: dict
        :param template_model:
        :type template_model: DataCollection
        """

        if not isinstance(from_dictionary, dict):
            raise TypeError("from_dictionary must be of type dict, %s \
                provided" % from_dictionary.__class__.__name__)

        rewrite_map = None
        if template_model is not None:

            if not isinstance(template_model, DataCollection):
                msg = "template_model should be a prestans model %s provided" % template_model.__class__.__name__
                raise TypeError(msg)

            rewrite_map = template_model.attribute_rewrite_reverse_map()

        for key, value in iter(from_dictionary.items()):

            target_key = key

            # minify support
            if rewrite_map is not None:
                target_key = rewrite_map[key]

            # ensure that the key exists in the template model
            if template_model is not None and target_key not in template_model:

                unwanted_keys = list()
                unwanted_keys.append(target_key)
                raise exception.AttributeFilterDiffers(unwanted_keys)

            # check to see we can work with the value
            if not isinstance(value, (bool, dict)):
                raise TypeError("AttributeFilter input for key %s must be \
                    boolean or dict, %s provided" % (key, value.__class__.__name__))

            # Either keep the value of wrap it up with AttributeFilter
            if isinstance(value, bool):
                setattr(self, target_key, value)
            elif isinstance(value, dict):

                sub_map = None
                if template_model is not None:

                    sub_map = getattr(template_model, target_key)

                    # prestans Array support
                    if isinstance(sub_map, Array):
                        sub_map = sub_map.element_template

                setattr(self, target_key, AttributeFilter(from_dictionary=value, template_model=sub_map))

    def __setattr__(self, key, value):
        """
        Overrides setattr to allow only booleans or an AttributeFilter
        """

        # Values should either be boolean or type of self
        if isinstance(value, bool) and key in self.__dict__ and isinstance(self.__dict__[key], self.__class__):
            self.__dict__[key].set_all_attribute_values(value)
            return
        elif isinstance(value, (bool, self.__class__)):
            self.__dict__[key] = value
            return

        raise TypeError("%s name in %s must be of type Boolean or AttributeFilter, given %s" %
                        (key, self.__class__.__name__, value.__class__.__name__))

    def as_immutable(self):
        """
        :return: immutable version of this attribute filter
        :rtype: AttributeFilterImmutable
        """
        from prestans.parser.attribute_filter_immutable import AttributeFilterImmutable
        return AttributeFilterImmutable(attribute_filter=self)



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

__all__ = ['VERB', 'STATUS']

#:
#: This file mostly contains constants for HTTP
#:


class VERB:
    """ 
    Encapsulates HTTP Verbs supported by the prestans framework in accordance
    with the REST definition. Each prestans REST handler must supporto at least
    one HTTP verb.

    HEAD is similar to a GET but it does not return a response.
    """

    GET = "GET"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    OPTIONS = "OPTIONS"

    @classmethod
    def is_supported_verb(cls, method):
        return method in [VERB.GET, VERB.HEAD, VERB.POST, VERB.PUT, VERB.PATCH, VERB.DELETE, VERB.OPTIONS]


class STATUS:
    """
    The following is a selection of HTTP status codes that are recommended for use 
    by REST services. You are welcome to use other available status codes.
    
    http://en.wikipedia.org/wiki/List_of_HTTP_status_codes
    """

    # Informational
    CONTINUE = 100
    SWITCHING_PROTOCOLS = 101
    PROCESSING = 102
    EARLY_HINTS = 103

    # Success
    OK = 200
    CREATED = 201
    ACCEPTED = 202
    NON_AUTHORITATIVE_INFORMATION = 203
    NO_CONTENT = 204
    RESET_CONTENT = 205
    PARTIAL_CONTENT = 206
    MULTI_STATUS = 207
    ALREADY_REPORTED = 208
    IM_USED = 226

    # Redirection
    MULTIPLE_CHOICES = 300
    MOVED_PERMANENTLY = 301
    FOUND = 302
    SEE_OTHER = 303
    NOT_MODIFIED = 304
    USE_PROXY = 305
    SWITCH_PROXY = 306
    TEMPORARY_REDIRECT = 307
    PERMANENT_REDIRECT = 308

    # Client Error
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    PAYMENT_REQUIRED = 402
    FORBIDDEN = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    NOT_ACCEPTABLE = 406
    PROXY_AUTH_REQUIRED = 407
    REQUEST_TIMEOUT = 408
    CONFLICT = 409
    GONE = 410
    LENGTH_REQUIRED = 411
    PRECONDITION_FAILED = 412
    PAYLOAD_TOO_LARGE = 413
    URI_TOO_LONG = 414
    UNSUPPORTED_MEDIA_TYPE = 415
    RANGE_NOT_SATISFIABLE = 416
    EXCEPTION_FAILED = 417
    TEAPOT = 418
    AUTHENTICATION_TIMEOUT = 419
    MISDIRECTED_REQUEST = 421
    UNPROCESSABLE_ENTITY = 422
    LOCKED = 423

    # Server Error
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    BAD_GATEWAY = 502
    SERVICE_UNAVAILABLE = 503
    GATEWAY_TIMEOUT = 504
    HTTP_VERSION_NOT_SUPPORTED = 505
    VARIANT_ALSO_NEGOTIATES = 506
    INSUFFICIENT_STORAGE = 507
    LOOP_DETECTED = 508
    NOT_EXTENDED = 510
    NETWORK_AUTHENTICATION_REQUIRED = 511



from prestans.http import VERB


class Config(object):
    """
    Configuration that's attached to each handler to define rules for each
    HTTP Verb. All HTTP verbs in use must have a configuration defined.

    __init__ takes in a VerbConfig instance for each parameter with names
    adjacent to the HTTP verb.
    """

    def __init__(self, GET=None, HEAD=None, POST=None, PUT=None, PATCH=None, DELETE=None, OPTIONS=None):
        from prestans.parser import VerbConfig

        for verb in [GET, HEAD, POST, PUT, PATCH, DELETE, OPTIONS]:
            if verb is not None and not isinstance(verb, VerbConfig):
                raise TypeError("All Parser configs should be of type prestans.parser.VerbConfig")

        self._configs = dict()

        self._configs[VERB.GET] = GET
        self._configs[VERB.HEAD] = HEAD
        self._configs[VERB.POST] = POST
        self._configs[VERB.PUT] = PUT
        self._configs[VERB.PATCH] = PATCH
        self._configs[VERB.DELETE] = DELETE
        self._configs[VERB.OPTIONS] = OPTIONS

    def get_config_for_verb(self, verb):
        return self._configs[verb]

    @property
    def get(self):
        return self._configs[VERB.GET]

    @property
    def head(self):
        return self._configs[VERB.HEAD]

    @property
    def post(self):
        return self._configs[VERB.POST]

    @property
    def put(self):
        return self._configs[VERB.PUT]

    @property
    def patch(self):
        return self._configs[VERB.PATCH]

    @property
    def delete(self):
        return self._configs[VERB.DELETE]

    @property
    def options(self):
        return self._configs[VERB.OPTIONS]



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from datetime import datetime
from datetime import time

from prestans import exception
from prestans.types import DataStructure


class Time(DataStructure):

    DEFAULT_FORMAT = "%H:%M:%S"
    NOW = '_PRESTANS_CONSTANT_MODEL_TIME_NOW'
    UTC_NOW = '_PRESTANS_CONSTANT_MODEL_TIME_UTC_NOW'

    def __init__(self, default=None, required=True, format=DEFAULT_FORMAT, description=None):

        if isinstance(default, time) or \
           default == self.NOW or \
           default == self.UTC_NOW or \
           default is None:
            self._default = default
        else:
            raise TypeError("default must be one of time, Time.NOW or Time.UTC_NOW")

        self._required = required
        self._format = format
        self._description = description

    @property
    def default(self):
        return self._default

    @property
    def required(self):
        """
        :return:
        :rtype: bool
        """
        return self._required

    @property
    def format(self):
        return self._format

    @property
    def description(self):
        return self._description

    def blueprint(self):

        blueprint = dict()
        blueprint['type'] = 'time'

        constraints = dict()
        constraints['default'] = self.default
        constraints['required'] = self.required
        constraints['format'] = self.format
        constraints['description'] = self.description

        blueprint['constraints'] = constraints
        return blueprint

    def validate(self, value):
        from prestans.util import string_types

        _validated_value = None

        # no need to do any validation if None, not required and default provided
        if not self._required and self._default is None and value is None:
            return value
        elif self._required and self._default is None and value is None:
            raise exception.RequiredAttributeError()
        elif value is None and self._default is not None:
            if self._default == Time.NOW:
                value = datetime.now()
            elif self._default == Time.UTC_NOW:
                value = datetime.utcnow()
            else:
                value = self._default

        if isinstance(value, time):
            _validated_value = value
        elif isinstance(value, string_types):
            try:
                _validated_value = datetime.strptime(value, self._format).time()
            except ValueError as exp:
                raise exception.ParseFailedError("time parsing failed %s" % exp)
        else:
            raise exception.ParseFailedError("cannot parse value of type %s" % value.__class__.__name__)

        return _validated_value

    def as_serializable(self, value):

        if not type(value) == time:
            raise exception.InvalidTypeError(value, 'datetime.time')

        return value.strftime(self._format)



# -*- coding: utf-8 -*-
#
#  prestans, A WSGI compliant REST micro-framework
#  http://prestans.org
#
#  Copyright (c) 2017, Anomaly Software Pty Ltd.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of Anomaly Software nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL ANOMALY SOFTWARE BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import re

from prestans import exception
from prestans.types import DataType


class String(DataType):

    def __init__(self, default=None, min_length=None, max_length=None,
                 required=True, format=None, choices=None, utf_encoding='utf-8',
                 description=None, trim=True):

        if min_length and max_length and min_length > max_length:
            raise ValueError("max length cannot be less than min length")

        if min_length and min_length < 1:
            raise ValueError("min length must be positive")

        if max_length and max_length < 1:
            raise ValueError("max length must be positive")

        self._default = default
        self._min_length = min_length
        self._max_length = max_length
        self._required = required
        self._format = format
        self._choices = choices
        self._utf_encoding = utf_encoding
        self._description = description
        self._trim = trim

    @property
    def required(self):
        return self._required

    @property
    def max_length(self):
        return self._max_length

    @property
    def min_length(self):
        return self._min_length

    @property
    def default(self):
        return self._default

    @property
    def choices(self):
        return self._choices

    @property
    def format(self):
        return self._format

    @property
    def description(self):
        return self._description

    @property
    def trim(self):
        return self._trim

    @property
    def utf_encoding(self):
        return self._utf_encoding

    def blueprint(self):

        blueprint = dict()
        blueprint['type'] = 'string'

        constraints = dict()
        constraints['default'] = self._default
        constraints['min_length'] = self._min_length
        constraints['max_length'] = self._max_length
        constraints['required'] = self._required
        constraints['format'] = self._format
        constraints['choices'] = self._choices
        constraints['utf_encoding'] = self._utf_encoding
        constraints['description'] = self._description
        constraints['trim'] = self._trim

        blueprint['constraints'] = constraints

        return blueprint

    def validate(self, value):

        _validated_value = None

        if self._required and self._default is None and value is None:
            raise exception.RequiredAttributeError()
        elif self._required and value is None:
            value = self._default
        elif not self._required and self._default is None and value is None:
            return _validated_value
        elif not self._required and value is None:
            value = self._default

        try:
            from prestans.util import string_types

            if isinstance(value, string_types):
                _validated_value = value
            else:
                _validated_value = str(value)
        except Exception as exp:
            raise exception.ParseFailedError("unicode or string encoding failed, %s" % exp)

        if self._trim:
            _validated_value = _validated_value.strip()

        # check for required and empty string
        if self._required and len(_validated_value) == 0:
            raise exception.RequiredAttributeError()

        if not self._required and len(_validated_value) == 0:
            return _validated_value

        if _validated_value is not None and self._min_length is not None and \
                len(_validated_value) < self._min_length:
            raise exception.MinimumLengthError(value, self._min_length)
        if _validated_value is not None and self._max_length is not None and \
                len(_validated_value) > self._max_length:
            raise exception.MaximumLengthError(value, self._max_length)

        if self._choices is not None and not _validated_value in self._choices:
            raise exception.InvalidChoiceError(value, self._choices)

        if self._format is not None and re.search(self._format, _validated_value) is None:
            raise exception.InvalidFormatError(_validated_value)

        return _validated_value
